from __future__ import annotations

import weakref
from abc import abstractmethod, ABC
from dataclasses import dataclass, field, replace
from enum import Enum, auto
from typing import List, Optional, TypeVar, Generic, ClassVar, Dict, Any, TYPE_CHECKING, Iterable, cast
from uuid import UUID

from rewrite import Markers
from rewrite import Tree, SourceFile, TreeVisitor

if TYPE_CHECKING:
    from .visitor import JavaVisitor

P = TypeVar('P')


class J(Tree):
    @property
    @abstractmethod
    def prefix(self) -> Space:
        ...

    def is_acceptable(self, v: TreeVisitor[Any, P], p: P) -> bool:
        from .visitor import JavaVisitor
        return isinstance(v, JavaVisitor)

    def accept(self, v: TreeVisitor[Any, P], p: P) -> Optional[Any]:
        from .visitor import JavaVisitor
        return self.accept_java(v.adapt(J, JavaVisitor), p)

    def accept_java(self, v: 'JavaVisitor[P]', p: P) -> Optional['J']:
        ...


@dataclass(frozen=True)
class Comment(ABC):
    @property
    @abstractmethod
    def multiline(self) -> bool:
        ...

    _text: str

    @property
    def text(self) -> str:
        return self._text

    _suffix: str

    @property
    def suffix(self) -> str:
        return self._suffix

    _markers: Markers

    @property
    def markers(self) -> Markers:
        return self._markers

    def replace(self, **kwargs) -> 'Comment':
        """Replace fields on this Comment using dataclasses.replace."""
        mapped_kwargs = {}
        for key, value in kwargs.items():
            if not key.startswith('_'):
                private_key = f'_{key}'
                if hasattr(self, private_key):
                    mapped_kwargs[private_key] = value
                else:
                    mapped_kwargs[key] = value
            else:
                mapped_kwargs[key] = value
        return replace(self, **mapped_kwargs)


@dataclass(frozen=True)
class TextComment(Comment):
    _multiline: bool

    @property
    def multiline(self) -> bool:
        return self._multiline

    # IMPORTANT: This explicit constructor aligns the parameter order with the Java side
    def __init__(self, _multiline: bool, _text: str, _suffix: str, _markers: Markers) -> None:
        object.__setattr__(self, '_multiline', _multiline)
        object.__setattr__(self, '_text', _text)
        object.__setattr__(self, '_suffix', _suffix)
        object.__setattr__(self, '_markers', _markers)


@dataclass(frozen=True)
class Space:
    _comments: List[Comment]

    @property
    def comments(self) -> List[Comment]:
        return self._comments

    _whitespace: Optional[str]

    @property
    def whitespace(self) -> str:
        return self._whitespace if self._whitespace is not None else ""

    def replace(self, **kwargs) -> 'Space':
        """Replace fields on this Space using dataclasses.replace."""
        mapped_kwargs = {}
        for key, value in kwargs.items():
            if not key.startswith('_'):
                private_key = f'_{key}'
                if hasattr(self, private_key):
                    mapped_kwargs[private_key] = value
                else:
                    mapped_kwargs[key] = value
            else:
                mapped_kwargs[key] = value
        return replace(self, **mapped_kwargs)

    def is_empty(self) -> bool:
        return len(self._comments) == 0 and (self._whitespace is None or self._whitespace == '')

    @classmethod
    def first_prefix(cls, trees: Optional[Iterable[J]]) -> Space:
        return Space.EMPTY if trees is None or not trees else next(iter(trees)).prefix

    @classmethod
    def format_first_prefix(cls, trees: List[J2], prefix: Space) -> List[J2]:
        if trees and next(iter(trees)).prefix != prefix:
            formatted_trees = list(trees)
            formatted_trees[0] = cast(J2, formatted_trees[0].replace(prefix=prefix))
            return formatted_trees
        return trees

    @property
    def indent(self) -> str:
        """
        The indentation after the last newline of either the last comment's suffix
        or the global whitespace if no comments exist.
        """
        return self._get_whitespace_indent(self.last_whitespace)

    @property
    def last_whitespace(self) -> str:
        """
        The raw suffix from the last comment if it exists, otherwise the global
        whitespace (or empty string if whitespace is None).
        """
        if self._comments:
            return self._comments[-1].suffix
        return self._whitespace if self._whitespace is not None else ""

    @staticmethod
    def _get_whitespace_indent(whitespace: Optional[str]) -> str:
        """
        A helper method that extracts everything after the last newline character
        in `whitespace`. If no newline is present, returns `whitespace` as-is.
        If the last newline is at the end, returns an empty string.
        """
        if not whitespace:
            return ""
        last_newline = whitespace.rfind('\n')
        return whitespace if last_newline == -1 else whitespace[last_newline + 1:]

    EMPTY: ClassVar[Space]
    SINGLE_SPACE: ClassVar[Space]

    class Location(Enum):
        ANNOTATED_TYPE_PREFIX = auto()
        ANNOTATIONS = auto()
        ANNOTATION_ARGUMENTS = auto()
        ANNOTATION_ARGUMENT_SUFFIX = auto()
        ANNOTATION_PREFIX = auto()
        ANY = auto()
        ARRAY_ACCESS_PREFIX = auto()
        ARRAY_INDEX_SUFFIX = auto()
        ARRAY_TYPE_PREFIX = auto()
        ASSERT_DETAIL = auto()
        ASSERT_DETAIL_PREFIX = auto()
        ASSERT_PREFIX = auto()
        ASSIGNMENT = auto()
        ASSIGNMENT_OPERATION_OPERATOR = auto()
        ASSIGNMENT_OPERATION_PREFIX = auto()
        ASSIGNMENT_PREFIX = auto()
        BINARY_OPERATOR = auto()
        BINARY_PREFIX = auto()
        BLOCK_END = auto()
        BLOCK_PREFIX = auto()
        BLOCK_STATEMENT_SUFFIX = auto()
        BREAK_PREFIX = auto()
        CASE = auto()
        CASE_BODY = auto()
        CASE_EXPRESSION = auto()
        CASE_PREFIX = auto()
        CASE_SUFFIX = auto()
        CATCH_ALTERNATIVE_SUFFIX = auto()
        CATCH_PREFIX = auto()
        CLASS_DECLARATION_PREFIX = auto()
        CLASS_KIND = auto()
        COMPILATION_UNIT_EOF = auto()
        COMPILATION_UNIT_PREFIX = auto()
        CONTINUE_PREFIX = auto()
        CONTROL_PARENTHESES_PREFIX = auto()
        DIMENSION = auto()
        DIMENSION_PREFIX = auto()
        DIMENSION_SUFFIX = auto()
        DO_WHILE_PREFIX = auto()
        ELSE_PREFIX = auto()
        EMPTY_PREFIX = auto()
        ENUM_VALUE_PREFIX = auto()
        ENUM_VALUE_SET_PREFIX = auto()
        ENUM_VALUE_SUFFIX = auto()
        ERRONEOUS_PREFIX = auto()
        EXPRESSION_PREFIX = auto()
        EXTENDS = auto()
        FIELD_ACCESS_NAME = auto()
        FIELD_ACCESS_PREFIX = auto()
        FOREACH_ITERABLE_SUFFIX = auto()
        FOREACH_VARIABLE_SUFFIX = auto()
        FOR_BODY_SUFFIX = auto()
        FOR_CONDITION_SUFFIX = auto()
        FOR_CONTROL_PREFIX = auto()
        FOR_EACH_CONTROL_PREFIX = auto()
        FOR_EACH_LOOP_PREFIX = auto()
        FOR_INIT_SUFFIX = auto()
        FOR_PREFIX = auto()
        FOR_UPDATE_SUFFIX = auto()
        IDENTIFIER_PREFIX = auto()
        IF_ELSE_SUFFIX = auto()
        IF_PREFIX = auto()
        IF_THEN_SUFFIX = auto()
        IMPLEMENTS = auto()
        IMPLEMENTS_SUFFIX = auto()
        IMPORT_ALIAS_PREFIX = auto()
        IMPORT_PREFIX = auto()
        IMPORT_SUFFIX = auto()
        INSTANCEOF_PREFIX = auto()
        INSTANCEOF_SUFFIX = auto()
        INTERSECTION_TYPE_PREFIX = auto()
        LABEL_PREFIX = auto()
        LABEL_SUFFIX = auto()
        LAMBDA_ARROW_PREFIX = auto()
        LAMBDA_PARAMETER = auto()
        LAMBDA_PARAMETERS_PREFIX = auto()
        LAMBDA_PREFIX = auto()
        LANGUAGE_EXTENSION = auto()
        LITERAL_PREFIX = auto()
        MEMBER_REFERENCE_CONTAINING = auto()
        MEMBER_REFERENCE_NAME = auto()
        MEMBER_REFERENCE_PREFIX = auto()
        METHOD_DECLARATION_DEFAULT_VALUE = auto()
        METHOD_DECLARATION_PARAMETERS = auto()
        METHOD_DECLARATION_PARAMETER_SUFFIX = auto()
        METHOD_DECLARATION_PREFIX = auto()
        METHOD_INVOCATION_ARGUMENTS = auto()
        METHOD_INVOCATION_ARGUMENT_SUFFIX = auto()
        METHOD_INVOCATION_NAME = auto()
        METHOD_INVOCATION_PREFIX = auto()
        METHOD_SELECT_SUFFIX = auto()
        MODIFIER_PREFIX = auto()
        MULTI_CATCH_PREFIX = auto()
        NAMED_VARIABLE_SUFFIX = auto()
        NEW_ARRAY_INITIALIZER = auto()
        NEW_ARRAY_INITIALIZER_SUFFIX = auto()
        NEW_ARRAY_PREFIX = auto()
        NEW_CLASS_ARGUMENTS = auto()
        NEW_CLASS_ARGUMENTS_SUFFIX = auto()
        NEW_CLASS_ENCLOSING_SUFFIX = auto()
        NEW_CLASS_PREFIX = auto()
        NEW_PREFIX = auto()
        NULLABLE_TYPE_PREFIX = auto()
        NULLABLE_TYPE_SUFFIX = auto()
        PACKAGE_PREFIX = auto()
        PACKAGE_SUFFIX = auto()
        PARAMETERIZED_TYPE_PREFIX = auto()
        PARENTHESES_PREFIX = auto()
        PARENTHESES_SUFFIX = auto()
        PERMITS = auto()
        PERMITS_SUFFIX = auto()
        PRIMITIVE_PREFIX = auto()
        RECORD_STATE_VECTOR = auto()
        RECORD_STATE_VECTOR_SUFFIX = auto()
        RETURN_PREFIX = auto()
        STATEMENT_PREFIX = auto()
        STATIC_IMPORT = auto()
        STATIC_INIT_SUFFIX = auto()
        SWITCH_EXPRESSION_PREFIX = auto()
        SWITCH_PREFIX = auto()
        SYNCHRONIZED_PREFIX = auto()
        TERNARY_FALSE = auto()
        TERNARY_PREFIX = auto()
        TERNARY_TRUE = auto()
        THROWS = auto()
        THROWS_SUFFIX = auto()
        THROW_PREFIX = auto()
        TRY_FINALLY = auto()
        TRY_PREFIX = auto()
        TRY_RESOURCE = auto()
        TRY_RESOURCES = auto()
        TRY_RESOURCE_SUFFIX = auto()
        TYPE_BOUNDS = auto()
        TYPE_BOUND_SUFFIX = auto()
        TYPE_CAST_PREFIX = auto()
        TYPE_PARAMETERS = auto()
        TYPE_PARAMETERS_PREFIX = auto()
        TYPE_PARAMETER_SUFFIX = auto()
        UNARY_OPERATOR = auto()
        UNARY_PREFIX = auto()
        UNKNOWN_PREFIX = auto()
        UNKNOWN_SOURCE_PREFIX = auto()
        VARARGS = auto()
        VARIABLE_DECLARATIONS_PREFIX = auto()
        VARIABLE_INITIALIZER = auto()
        VARIABLE_PREFIX = auto()
        WHILE_BODY_SUFFIX = auto()
        WHILE_CONDITION = auto()
        WHILE_PREFIX = auto()
        WILDCARD_BOUND = auto()
        WILDCARD_PREFIX = auto()
        YIELD_PREFIX = auto()


Space.EMPTY = Space([], '')
Space.SINGLE_SPACE = Space([], ' ')


@dataclass
class CoordinateBuilder:
    tree: J

    def before(self, loc: Space.Location) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc, JavaCoordinates.Mode.BEFORE)

    def after(self, loc: Space.Location) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc, JavaCoordinates.Mode.AFTER)

    def replace(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc, JavaCoordinates.Mode.REPLACE)


@dataclass
class _ExpressionCoordinateBuilder(CoordinateBuilder):
    def after(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc or Space.Location.EXPRESSION_PREFIX, JavaCoordinates.Mode.AFTER)

    def before(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc or Space.Location.EXPRESSION_PREFIX, JavaCoordinates.Mode.BEFORE)

    def replace(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc or Space.Location.EXPRESSION_PREFIX, JavaCoordinates.Mode.REPLACE)


@dataclass
class _StatementCoordinateBuilder(CoordinateBuilder):
    def after(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc or Space.Location.STATEMENT_PREFIX, JavaCoordinates.Mode.AFTER)

    def before(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc or Space.Location.STATEMENT_PREFIX, JavaCoordinates.Mode.BEFORE)

    def replace(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
        return JavaCoordinates(self.tree, loc or Space.Location.STATEMENT_PREFIX, JavaCoordinates.Mode.REPLACE)


@dataclass
class _BlockCoordinateBuilder(_StatementCoordinateBuilder):
    def first_statement(self) -> JavaCoordinates:
        if not self.tree.statements:
            return self.last_statement()
        return self.tree.statements[0].get_coordinates().before()

    def last_statement(self) -> JavaCoordinates:
        return self.before(Space.Location.BLOCK_END)


CoordinateBuilder.Expression = _ExpressionCoordinateBuilder  # type: ignore
CoordinateBuilder.Statement = _StatementCoordinateBuilder  # type: ignore
CoordinateBuilder.Block = _BlockCoordinateBuilder  # type: ignore


@dataclass
class JavaCoordinates:
    tree: J
    loc: Space.Location
    mode: Mode

    def is_replacement(self) -> bool:
        return self.mode == JavaCoordinates.Mode.REPLACE

    class Mode(Enum):
        AFTER = 0,
        BEFORE = 1,
        REPLACE = 2,


class JavaSourceFile(J, SourceFile):
    pass


class Expression(J):
    def get_coordinates(self) -> CoordinateBuilder.Expression:  # type: ignore
        return CoordinateBuilder.Expression(self)  # type: ignore


class Statement(J):
    def get_coordinates(self) -> CoordinateBuilder.Statement:  # type: ignore
        return CoordinateBuilder.Statement(self)  # type: ignore


class TypedTree(J):
    pass


class NameTree(TypedTree):
    pass


class TypeTree(NameTree):
    pass


class Loop(Statement):
    pass


class MethodCall(Expression):
    pass


class JavaType(ABC):
    class FullyQualified:
        class Kind(Enum):
            Class = 0
            Enum = 1
            Interface = 2
            Annotation = 3
            Record = 4

    class Unknown(FullyQualified):
        pass

    class Class(FullyQualified):
        _flags_bit_map: int
        _fully_qualified_name: str
        _kind: FullyQualified.Kind
        _type_parameters: Optional[List[JavaType]]
        _supertype: Optional[JavaType.FullyQualified]
        _owning_class: Optional[JavaType.FullyQualified]
        _annotations: Optional[List[JavaType.FullyQualified]]
        _interfaces: Optional[List[JavaType.FullyQualified]]
        _members: Optional[List[JavaType.Variable]]
        _methods: Optional[List[JavaType.Method]]

    class ShallowClass(Class):
        pass

    class Parameterized(FullyQualified):
        _type: JavaType.FullyQualified
        _type_parameters: Optional[List[JavaType]]

    class GenericTypeVariable:
        class Variance(Enum):
            Invariant = 0
            Covariant = 1
            Contravariant = 2

    class Primitive(Enum):
        Boolean = 0
        Byte = 1
        Char = 2
        Double = 3
        Float = 4
        Int = 5
        Long = 6
        Short = 7
        Void = 8
        String = 9
        None_ = 10
        Null = 11

        @classmethod
        def _missing_(cls, value):
            if value is None:
                return cls.None_
            return super()._missing_(value)

    @dataclass
    class Method:
        _flags_bit_map: int = field(default=0)
        _declaring_type: Optional[JavaType.FullyQualified] = field(default=None)
        _name: str = field(default="")
        _return_type: Optional[JavaType] = field(default=None)
        _parameter_names: Optional[List[str]] = field(default=None)
        _parameter_types: Optional[List[JavaType]] = field(default=None)
        _thrown_exceptions: Optional[List[JavaType]] = field(default=None)
        _annotations: Optional[List[JavaType.FullyQualified]] = field(default=None)
        _default_value: Optional[List[str]] = field(default=None)
        _declared_formal_type_names: Optional[List[str]] = field(default=None)

        @property
        def flags_bit_map(self) -> int:
            return self._flags_bit_map

        @property
        def declaring_type(self) -> Optional[JavaType.FullyQualified]:
            return self._declaring_type

        @property
        def name(self) -> str:
            return self._name

        @property
        def return_type(self) -> Optional[JavaType]:
            return self._return_type

        @property
        def parameter_names(self) -> Optional[List[str]]:
            return self._parameter_names

        @property
        def parameter_types(self) -> Optional[List[JavaType]]:
            return self._parameter_types

        @property
        def thrown_exceptions(self) -> Optional[List[JavaType]]:
            return self._thrown_exceptions

        @property
        def annotations(self) -> Optional[List[JavaType.FullyQualified]]:
            return self._annotations

        @property
        def default_value(self) -> Optional[List[str]]:
            return self._default_value

        @property
        def declared_formal_type_names(self) -> Optional[List[str]]:
            return self._declared_formal_type_names

    class Variable:
        pass

    class Array:
        pass


T = TypeVar('T')
J2 = TypeVar('J2', bound=J)
J3 = TypeVar('J3', bound=J)


@dataclass(frozen=True)
class JRightPadded(Generic[T]):
    _element: T

    @property
    def element(self) -> T:
        return self._element

    _after: Space

    @property
    def after(self) -> Space:
        return self._after

    _markers: Markers

    @property
    def markers(self) -> Markers:
        return self._markers

    def __eq__(self, other) -> bool:
        if isinstance(other, JRightPadded):
            return self._element == other._element
        return self._element == other

    def __hash__(self) -> int:
        return hash(self._element)

    def replace(self, **kwargs) -> 'JRightPadded[T]':
        """Replace fields using keyword arguments."""
        # Map public property names to private field names
        mapped = {}
        for key, value in kwargs.items():
            if not key.startswith('_') and hasattr(self, f'_{key}'):
                mapped[f'_{key}'] = value
            else:
                mapped[key] = value
        return replace(self, **mapped)

    @classmethod
    def get_elements(cls, padded_list: List[JRightPadded[T]]) -> List[T]:
        return [x.element for x in padded_list]

    @classmethod
    def merge_elements(cls, before: List[JRightPadded[J2]], elements: List[J2]) -> List[JRightPadded[J2]]:
        # a cheaper check for the most common case when there are no changes
        if len(elements) == len(before):
            has_changes = False
            for i in range(len(before)):
                if before[i].element is not elements[i]:
                    has_changes = True
                    break
            if not has_changes:
                return before
        elif not elements:
            return []

        after: List[JRightPadded[J2]] = []
        before_by_id: Dict[UUID, JRightPadded[J2]] = {}

        for j in before:
            if j.element.id in before_by_id:
                raise Exception("Duplicate key")
            before_by_id[j.element.id] = j

        for t in elements:
            found = before_by_id.get(t.id)
            if found is not None:
                after.append(found.replace(element=t))
            else:
                after.append(JRightPadded(t, Space.EMPTY, Markers.EMPTY))

        return after

    class Location(Enum):
        ANNOTATION_ARGUMENT = Space.Location.ANNOTATION_ARGUMENT_SUFFIX
        ARRAY_INDEX = Space.Location.ARRAY_INDEX_SUFFIX
        BLOCK_STATEMENT = Space.Location.BLOCK_STATEMENT_SUFFIX
        CASE = Space.Location.CASE_SUFFIX
        CASE_BODY = Space.Location.CASE_BODY
        CASE_EXPRESSION = Space.Location.CASE_EXPRESSION
        CATCH_ALTERNATIVE = Space.Location.CATCH_ALTERNATIVE_SUFFIX
        DIMENSION = Space.Location.DIMENSION_SUFFIX
        ENUM_VALUE = Space.Location.ENUM_VALUE_SUFFIX
        FOREACH_ITERABLE = Space.Location.FOREACH_ITERABLE_SUFFIX
        FOREACH_VARIABLE = Space.Location.FOREACH_VARIABLE_SUFFIX
        FOR_BODY = Space.Location.FOR_BODY_SUFFIX
        FOR_CONDITION = Space.Location.FOR_CONDITION_SUFFIX
        FOR_INIT = Space.Location.FOR_INIT_SUFFIX
        FOR_UPDATE = Space.Location.FOR_UPDATE_SUFFIX
        IF_ELSE = Space.Location.IF_ELSE_SUFFIX
        IF_THEN = Space.Location.IF_THEN_SUFFIX
        IMPLEMENTS = Space.Location.IMPLEMENTS_SUFFIX
        IMPORT = Space.Location.IMPORT_SUFFIX
        INSTANCEOF = Space.Location.INSTANCEOF_SUFFIX
        LABEL = Space.Location.LABEL_SUFFIX
        LAMBDA_PARAM = Space.Location.LAMBDA_PARAMETER
        LANGUAGE_EXTENSION = Space.Location.LANGUAGE_EXTENSION
        MEMBER_REFERENCE_CONTAINING = Space.Location.MEMBER_REFERENCE_CONTAINING
        METHOD_DECLARATION_PARAMETER = Space.Location.METHOD_DECLARATION_PARAMETER_SUFFIX
        METHOD_INVOCATION_ARGUMENT = Space.Location.METHOD_INVOCATION_ARGUMENT_SUFFIX
        METHOD_SELECT = Space.Location.METHOD_SELECT_SUFFIX
        NAMED_VARIABLE = Space.Location.NAMED_VARIABLE_SUFFIX
        NEW_ARRAY_INITIALIZER = Space.Location.NEW_ARRAY_INITIALIZER_SUFFIX
        NEW_CLASS_ARGUMENTS = Space.Location.NEW_CLASS_ARGUMENTS_SUFFIX
        NEW_CLASS_ENCLOSING = Space.Location.NEW_CLASS_ENCLOSING_SUFFIX
        NULLABLE = Space.Location.NULLABLE_TYPE_SUFFIX
        PACKAGE = Space.Location.PACKAGE_SUFFIX
        PARENTHESES = Space.Location.PARENTHESES_SUFFIX
        PERMITS = Space.Location.PERMITS_SUFFIX
        RECORD_STATE_VECTOR = Space.Location.RECORD_STATE_VECTOR_SUFFIX
        STATIC_INIT = Space.Location.STATIC_INIT_SUFFIX
        THROWS = Space.Location.THROWS_SUFFIX
        TRY_RESOURCE = Space.Location.TRY_RESOURCE_SUFFIX
        TYPE_BOUND = Space.Location.TYPE_BOUND_SUFFIX
        TYPE_PARAMETER = Space.Location.TYPE_PARAMETER_SUFFIX
        WHILE_BODY = Space.Location.WHILE_BODY_SUFFIX

        def __init__(self, after_location: Space.Location):
            self.after_location = after_location


@dataclass(frozen=True)
class JLeftPadded(Generic[T]):
    _before: Space

    @property
    def before(self) -> Space:
        return self._before

    _element: T

    @property
    def element(self) -> T:
        return self._element

    _markers: Markers

    @property
    def markers(self) -> Markers:
        return self._markers

    def __eq__(self, other) -> bool:
        if isinstance(other, JLeftPadded):
            return self._element == other._element
        return self._element == other

    def __hash__(self) -> int:
        return hash(self._element)

    def replace(self, **kwargs) -> 'JLeftPadded[T]':
        """Replace fields using keyword arguments."""
        # Map public property names to private field names
        mapped = {}
        for key, value in kwargs.items():
            if not key.startswith('_') and hasattr(self, f'_{key}'):
                mapped[f'_{key}'] = value
            else:
                mapped[key] = value
        return replace(self, **mapped)

    class Location(Enum):
        ARRAY_TYPE_DIMENSION = Space.Location.DIMENSION_PREFIX
        ASSERT_DETAIL = Space.Location.ASSERT_DETAIL_PREFIX
        ASSIGNMENT = Space.Location.ASSIGNMENT
        ASSIGNMENT_OPERATION_OPERATOR = Space.Location.ASSIGNMENT_OPERATION_OPERATOR
        BINARY_OPERATOR = Space.Location.BINARY_OPERATOR
        CLASS_KIND = Space.Location.CLASS_KIND
        DIMENSION_PREFIX = Space.Location.CLASS_KIND
        EXTENDS = Space.Location.EXTENDS
        FIELD_ACCESS_NAME = Space.Location.FIELD_ACCESS_NAME
        IMPORT_ALIAS_PREFIX = Space.Location.IMPORT_ALIAS_PREFIX
        LANGUAGE_EXTENSION = Space.Location.LANGUAGE_EXTENSION
        MEMBER_REFERENCE_NAME = Space.Location.MEMBER_REFERENCE_NAME
        METHOD_DECLARATION_DEFAULT_VALUE = Space.Location.METHOD_DECLARATION_DEFAULT_VALUE
        STATIC_IMPORT = Space.Location.STATIC_IMPORT
        TERNARY_FALSE = Space.Location.TERNARY_FALSE
        TERNARY_TRUE = Space.Location.TERNARY_TRUE
        TRY_FINALLY = Space.Location.TRY_FINALLY
        UNARY_OPERATOR = Space.Location.UNARY_OPERATOR
        VARIABLE_INITIALIZER = Space.Location.VARIABLE_INITIALIZER
        WHILE_CONDITION = Space.Location.WHILE_CONDITION
        WILDCARD_BOUND = Space.Location.WILDCARD_BOUND

        def __init__(self, before_location: Space.Location):
            self.before_location = before_location


@dataclass(frozen=True)
class JContainer(Generic[J2]):
    _before: Space

    @property
    def before(self) -> Space:
        return self._before

    _elements: List[JRightPadded[J2]]

    @property
    def elements(self) -> List[J2]:
        return JRightPadded.get_elements(self._elements)

    _markers: Markers

    @property
    def markers(self) -> Markers:
        return self._markers

    def __eq__(self, other) -> bool:
        if isinstance(other, JContainer):
            return self._elements == other._elements
        if isinstance(other, list):
            return self.elements == other
        return False

    def __hash__(self) -> int:
        return hash(tuple(self._elements))

    def replace(self, **kwargs) -> 'JContainer[J2]':
        """Replace fields using keyword arguments."""
        # Map public property names to private field names
        mapped = {}
        for key, value in kwargs.items():
            if not key.startswith('_') and hasattr(self, f'_{key}'):
                mapped[f'_{key}'] = value
            else:
                mapped[key] = value
        return replace(self, **mapped)

    @dataclass
    class PaddingHelper(Generic[J3]):
        _t: JContainer[J3]

        @property
        def elements(self) -> List[JRightPadded[J3]]:
            return self._t._elements

        def replace(self, **kwargs) -> JContainer[J3]:
            """Replace fields of the container using keyword arguments."""
            if 'elements' in kwargs:
                elements = kwargs['elements']
                if self._t._elements is elements:
                    return self._t
                return JContainer(self._t._before, elements, self._t._markers)
            return self._t

    _padding: Optional[weakref.ReferenceType[JContainer.PaddingHelper[J2]]] = None

    @property
    def padding(self) -> JContainer.PaddingHelper[J2]:
        p: Optional[JContainer.PaddingHelper[J2]]
        if self._padding is None:
            p = JContainer.PaddingHelper(self)
            object.__setattr__(self, '_padding', weakref.ref(p))
        else:
            p = self._padding()
            # noinspection PyProtectedMember
            if p is None or p._t != self:
                p = JContainer.PaddingHelper(self)
                object.__setattr__(self, '_padding', weakref.ref(p))
        return p

    @classmethod
    def build_nullable(cls, before: Optional[JContainer[J2]], elements: Optional[List[J2]]) -> Optional[
        JContainer[J2]]:
        if elements is None or elements == []:
            return None
        if before is None:
            return JContainer(Space.EMPTY, JRightPadded.merge_elements([], elements), Markers.EMPTY)
        return before.padding.replace(elements=JRightPadded.merge_elements(before._elements, elements))

    _EMPTY: Optional[JContainer[J]] = None

    @classmethod
    def empty(cls) -> JContainer[J2]:
        if cls._EMPTY is None:
            cls._EMPTY = JContainer(Space.EMPTY, [], Markers.EMPTY)
        return cls._EMPTY  # type: ignore

    class Location(Enum):
        ANNOTATION_ARGUMENTS = (Space.Location.ANNOTATION_ARGUMENTS, JRightPadded.Location.ANNOTATION_ARGUMENT)
        CASE = (Space.Location.CASE, JRightPadded.Location.CASE)
        CASE_EXPRESSION = (Space.Location.CASE_EXPRESSION, JRightPadded.Location.CASE_EXPRESSION)
        IMPLEMENTS = (Space.Location.IMPLEMENTS, JRightPadded.Location.IMPLEMENTS)
        LANGUAGE_EXTENSION = (Space.Location.LANGUAGE_EXTENSION, JRightPadded.Location.LANGUAGE_EXTENSION)
        METHOD_DECLARATION_PARAMETERS = (
            Space.Location.METHOD_DECLARATION_PARAMETERS, JRightPadded.Location.METHOD_DECLARATION_PARAMETER)
        METHOD_INVOCATION_ARGUMENTS = (
            Space.Location.METHOD_INVOCATION_ARGUMENTS, JRightPadded.Location.METHOD_INVOCATION_ARGUMENT)
        NEW_ARRAY_INITIALIZER = (Space.Location.NEW_ARRAY_INITIALIZER, JRightPadded.Location.NEW_ARRAY_INITIALIZER)
        NEW_CLASS_ARGUMENTS = (Space.Location.NEW_CLASS_ARGUMENTS, JRightPadded.Location.NEW_CLASS_ARGUMENTS)
        PERMITS = (Space.Location.PERMITS, JRightPadded.Location.PERMITS)
        RECORD_STATE_VECTOR = (Space.Location.RECORD_STATE_VECTOR, JRightPadded.Location.RECORD_STATE_VECTOR)
        THROWS = (Space.Location.THROWS, JRightPadded.Location.THROWS)
        TRY_RESOURCES = (Space.Location.TRY_RESOURCES, JRightPadded.Location.TRY_RESOURCE)
        TYPE_BOUNDS = (Space.Location.TYPE_BOUNDS, JRightPadded.Location.TYPE_BOUND)
        TYPE_PARAMETERS = (Space.Location.TYPE_PARAMETERS, JRightPadded.Location.TYPE_PARAMETER)

        def __init__(self, before_location: Space.Location, element_location: JRightPadded.Location):
            self.before_location = before_location
            self.element_location = element_location



# Auto-generated stub file for IDE autocomplete support.
# Do not edit manually - regenerate with: python scripts/generate_stubs.py

from typing import Any, List, Optional, Self
from uuid import UUID


class Markers:
    id: UUID
    markers: List[Marker]

    def replace(
        self,
        *,
        id: UUID = ...,
        markers: List[Marker] = ...,
    ) -> Self: ...

class SearchResult(Marker):
    id: UUID
    description: Optional[str]

    def replace(
        self,
        *,
        id: UUID = ...,
        description: Optional[str] = ...,
    ) -> Self: ...

class ParseExceptionResult(Marker):
    id: UUID
    parser_type: str
    exception_type: str
    message: str

    def replace(
        self,
        *,
        id: UUID = ...,
        parser_type: str = ...,
        exception_type: str = ...,
        message: str = ...,
    ) -> Self: ...

class UnknownJavaMarker(Marker):
    id: UUID
    data: Dict[str, Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        data: Dict[str, Any] = ...,
    ) -> Self: ...



# Copyright 2025 the original author or authors.
# <p>
# Licensed under the Moderne Source Available License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# <p>
# https://docs.moderne.io/licensing/moderne-source-available-license
# <p>
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Recipe to remove redundant pass statements from Python code."""

from typing import Any

from rewrite import ExecutionContext, Recipe, TreeVisitor
from rewrite.category import CategoryDescriptor
from rewrite.decorators import categorize
from rewrite.marketplace import Python
from rewrite.python.tree import Pass
from rewrite.python.visitor import PythonVisitor
from rewrite.java import J

# Define category path locally to avoid circular imports
_Cleanup = [*Python, CategoryDescriptor(display_name="Cleanup")]


@categorize(_Cleanup)
class RemovePass(Recipe):
    """
    Remove `pass` statements from Python code.

    This recipe removes all `pass` statements. Note that this may make
    code syntactically invalid if the `pass` is the only statement in a
    block (e.g., empty function body).
    """

    @property
    def name(self) -> str:
        return "org.openrewrite.python.RemovePass"

    @property
    def display_name(self) -> str:
        return "Remove pass statements"

    @property
    def description(self) -> str:
        return "Remove `pass` statements from Python code."

    def editor(self) -> TreeVisitor[Any, ExecutionContext]:
        class Visitor(PythonVisitor[ExecutionContext]):
            def visit_pass(self, pass_: Pass, ctx: ExecutionContext) -> J:
                return None

        return Visitor()



# Auto-generated stub file for IDE autocomplete support.
# Do not edit manually - regenerate with: python scripts/generate_stubs.py

from typing import Optional, Self

from rewrite.markers import Marker


class FileAttributes:
    creation_time: Optional[datetime]
    last_modified_time: Optional[datetime]
    last_access_time: Optional[datetime]
    is_readable: bool
    is_writable: bool
    is_executable: bool
    size: int

    def replace(
        self,
        *,
        creation_time: Optional[datetime] = ...,
        last_modified_time: Optional[datetime] = ...,
        last_access_time: Optional[datetime] = ...,
        is_readable: bool = ...,
        is_writable: bool = ...,
        is_executable: bool = ...,
        size: int = ...,
    ) -> Self: ...

class Checksum:
    algorithm: str
    value: bytes

    def replace(
        self,
        *,
        algorithm: str = ...,
        value: bytes = ...,
    ) -> Self: ...



import io
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from pathlib import Path
from time import time_ns
from typing import Iterable, Optional, Callable, TypeVar, Any, cast, IO
from uuid import UUID

from .execution import ExecutionContext, InMemoryExecutionContext
from .markers import Markers, ParseExceptionResult
from .result import Result
from .tree import SourceFile, Tree, PrintOutputCapture, PrinterFactory, FileAttributes, Checksum
from .utils import random_id
from .visitor import TreeVisitor, Cursor


@dataclass(frozen=True, eq=False)
class ParserInput:
    _path: Path

    @property
    def path(self) -> Path:
        return self._path

    _file_attributes: Optional[FileAttributes]

    @property
    def file_attributes(self) -> Optional[FileAttributes]:
        return self._file_attributes

    _synthetic: bool

    @property
    def synthetic(self) -> bool:
        return self._synthetic

    _source: Callable[[], IO[Any]]

    @property
    def source(self) -> Callable[[], IO[Any]]:
        return self._source


P = TypeVar('P')


# noinspection PyShadowingBuiltins,PyShadowingNames,DuplicatedCode
@dataclass(frozen=True, eq=False)
class ParseError(SourceFile):
    @classmethod
    def build(cls, parser: 'Parser', input: ParserInput, relative_to: Optional[Path], ctx: ExecutionContext, exception: Exception,
              erroneous: Optional[SourceFile] = None) -> 'ParseError':
        return cls(random_id(),
                   Markers(random_id(), [ParseExceptionResult.build(parser, exception)]),
                   input.path.relative_to(relative_to) if relative_to else input.path,
                   input.file_attributes, parser.get_charset(ctx), False,
                   None,
                   input.source().read(),
                   erroneous)

    _id: UUID

    @property
    def id(self) -> UUID:
        return self._id

    def with_id(self, id: UUID) -> 'ParseError':
        return self if id is self._id else replace(self, _id=id)

    _markers: Markers

    @property
    def markers(self) -> Markers:
        return self._markers

    def with_markers(self, markers: Markers) -> 'ParseError':
        return self if markers is self._markers else replace(self, _markers=markers)

    _source_path: Path

    @property
    def source_path(self) -> Path:
        return self._source_path

    def with_source_path(self, source_path: Path) -> 'ParseError':
        return self if source_path is self._source_path else replace(self, _source_path=source_path)

    _file_attributes: Optional[FileAttributes]

    @property
    def file_attributes(self) -> Optional[FileAttributes]:
        return self._file_attributes

    def with_file_attributes(self, file_attributes: Optional[FileAttributes]) -> 'ParseError':
        return self if file_attributes is self._file_attributes else replace(self, _file_attributes=file_attributes)

    _charset_name: Optional[str]

    @property
    def charset_name(self) -> Optional[str]:
        return self._charset_name

    def with_charset_name(self, charset_name: Optional[str]) -> 'ParseError':
        return self if charset_name is self._charset_name else replace(self, _charset_name=charset_name)

    _charset_bom_marked: bool

    @property
    def charset_bom_marked(self) -> bool:
        return self._charset_bom_marked

    def with_charset_bom_marked(self, charset_bom_marked: bool) -> 'ParseError':
        return self if charset_bom_marked is self._charset_bom_marked else replace(self,
                                                                                   _charset_bom_marked=charset_bom_marked)

    _checksum: Optional[Checksum]

    @property
    def checksum(self) -> Optional[Checksum]:
        return self._checksum

    def with_checksum(self, checksum: Optional[Checksum]) -> 'ParseError':
        return self if checksum is self._checksum else replace(self, _checksum=checksum)

    _text: str

    @property
    def text(self) -> str:
        return self._text

    def with_text(self, text: str) -> 'ParseError':
        return self if text is self._text else replace(self, _text=text)

    _erroneous: Optional[SourceFile]

    @property
    def erroneous(self) -> Optional[SourceFile]:
        return self._erroneous

    def with_erroneous(self, erroneous: Optional[SourceFile]) -> 'ParseError':
        return self if erroneous is self._erroneous else replace(self, _erroneous=erroneous)

    def printer(self, cursor: Cursor) -> TreeVisitor[Tree, PrintOutputCapture[P]]:
        return PrinterFactory.current().create_printer(cursor)

    def is_acceptable(self, v: TreeVisitor[Any, P], p: P) -> bool:
        return isinstance(v, ParseErrorVisitor)

    def accept(self, v: TreeVisitor[Any, P], p: P) -> Optional[Any]:
        return cast(ParseErrorVisitor, v).visit_parse_error(self, p)


class ParseErrorVisitor(TreeVisitor[Tree, P]):
    def is_acceptable(self, source_file: SourceFile, p: P) -> bool:
        return isinstance(source_file, ParseError)

    def visit_parse_error(self, e: ParseError, p: P) -> ParseError:
        return e.replace(markers=self.visit_markers(e.markers, p))


class Parser(ABC):
    @abstractmethod
    def parse_inputs(self, sources: Iterable[ParserInput], relative_to: Optional[Path],
                     ctx: ExecutionContext) -> Iterable[SourceFile]:
        pass

    @abstractmethod
    def accept(self, path: Path) -> bool:
        pass

    @abstractmethod
    def source_path_from_source_text(self, prefix: Path, source_code: str) -> Path:
        pass

    def parse(self, source_files: Iterable[Path], relative_to: Optional[Path], ctx: ExecutionContext) -> Iterable[
        SourceFile]:
        inputs = [ParserInput(path, None, False, lambda: io.FileIO(path)) for path in source_files]
        return self.parse_inputs(inputs, relative_to, ctx)

    def parse_strings(self, *sources: str) -> Iterable[SourceFile]:
        return self.__parse_strings_with_context(InMemoryExecutionContext(), *sources)

    def __parse_strings_with_context(self, ctx: ExecutionContext, *sources: str) -> Iterable[SourceFile]:
        inputs = [ParserInput(self.source_path_from_source_text(Path(str(time_ns())), source),
                              None,
                              True,
                              lambda: io.StringIO(source)) for source in sources]
        return self.parse_inputs(inputs, None, ctx)

    def accept_input(self, parser_input: ParserInput) -> bool:
        return parser_input.synthetic or self.accept(parser_input.path)

    def accepted_inputs(self, inputs: Iterable[ParserInput]) -> Iterable[ParserInput]:
        return filter(self.accept_input, inputs)

    def reset(self):
        return self

    def get_charset(self, ctx: ExecutionContext) -> str:
        return cast(str, ctx.get_message(ExecutionContext.CHARSET, 'utf-8'))


class ParserBuilder(ABC):
    _source_file_type: type

    def __init__(self, source_file_type: type):
        self._source_file_type = source_file_type

    @property
    def source_file_type(self) -> type:
        return self._source_file_type

    _dsl_name: Optional[str]

    @property
    def dsl_name(self) -> Optional[str]:
        return self._dsl_name

    @abstractmethod
    def build(self) -> Parser:
        pass


def require_print_equals_input(parser: Parser, source_file: SourceFile, parser_input: ParserInput,
                               relative_to: Optional[Path], ctx: ExecutionContext) -> SourceFile:
    required = ctx.get_message(ExecutionContext.REQUIRE_PRINT_EQUALS_INPUT, True)
    if (required and not source_file.print_equals_input(parser_input, ctx)):
        diff = Result.diff(
            parser_input.source().read(),
            source_file.print_all(),
            parser_input.path
        )
        return (ParseError.build(parser, parser_input, relative_to, ctx,
                                 Exception(f"{source_file.source_path} is not print idempotent. \n{diff}"),
                                 source_file))
    return source_file



# Auto-generated stub file for IDE autocomplete support.
# Do not edit manually - regenerate with: python scripts/generate_stubs.py

from pathlib import Path
from typing import Any, List, Optional, Self
from uuid import UUID

from rewrite.java.support_types import J, Expression, Statement, TypedTree, TypeTree, NameTree, Loop, Space, \
    JRightPadded, JLeftPadded, JContainer
from rewrite.markers import Markers
from rewrite.python.support_types import Py
from rewrite.tree import Checksum, FileAttributes, SourceFile


class Async(Py, Statement):
    id: UUID
    prefix: Space
    markers: Markers
    statement: Statement

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        statement: Statement = ...,
    ) -> Self: ...

class Await(Py, Expression):
    id: UUID
    prefix: Space
    markers: Markers
    expression: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        expression: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class Binary(Py, Expression, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    left: Expression
    operator: JLeftPadded[Type]
    negation: Optional[Space]
    right: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        left: Expression = ...,
        operator: JLeftPadded[Type] = ...,
        negation: Optional[Space] = ...,
        right: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class ChainedAssignment(Py, Statement, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    variables: List[JRightPadded[Expression]]
    assignment: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        variables: List[JRightPadded[Expression]] = ...,
        assignment: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class CollectionLiteral(Py, Expression, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    collection_kind: Kind
    elements: JContainer[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        collection_kind: Kind = ...,
        elements: JContainer[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class CompilationUnit(Py, SourceFile):
    id: UUID
    prefix: Space
    markers: Markers
    source_path: Path
    charset_bom_marked: bool
    imports: List[JRightPadded[Statement]]
    statements: List[JRightPadded[Statement]]
    eof: Space
    file_attributes: Optional[FileAttributes]
    charset_name: Optional[str]
    checksum: Optional[Checksum]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        source_path: Path = ...,
        charset_bom_marked: bool = ...,
        imports: List[JRightPadded[Statement]] = ...,
        statements: List[JRightPadded[Statement]] = ...,
        eof: Space = ...,
        file_attributes: Optional[FileAttributes] = ...,
        charset_name: Optional[str] = ...,
        checksum: Optional[Checksum] = ...,
    ) -> Self: ...

class ComprehensionExpression(Py, Expression):
    class Condition:
        id: UUID
        prefix: Space
        markers: Markers
        expression: Expression

        def replace(
            self,
            *,
            id: UUID = ...,
            prefix: Space = ...,
            markers: Markers = ...,
            expression: Expression = ...,
        ) -> Self: ...

    class Clause:
        id: UUID
        prefix: Space
        markers: Markers
        iterator_variable: Expression
        iterated_list: JLeftPadded[Expression]
        conditions: List[Condition]

        def replace(
            self,
            *,
            id: UUID = ...,
            prefix: Space = ...,
            markers: Markers = ...,
            iterator_variable: Expression = ...,
            iterated_list: JLeftPadded[Expression] = ...,
            conditions: List[Condition] = ...,
        ) -> Self: ...

    id: UUID
    prefix: Space
    markers: Markers
    comprehension_kind: Kind
    result: Expression
    clauses: List[Clause]
    suffix: Space
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        comprehension_kind: Kind = ...,
        result: Expression = ...,
        clauses: List[Clause] = ...,
        suffix: Space = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class Del(Py, Statement):
    id: UUID
    prefix: Space
    markers: Markers
    targets: List[JRightPadded[Expression]]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        targets: List[JRightPadded[Expression]] = ...,
    ) -> Self: ...

class DictLiteral(Py, Expression, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    elements: JContainer[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        elements: JContainer[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class ErrorFrom(Py, Expression):
    id: UUID
    prefix: Space
    markers: Markers
    error: Expression
    from_: JLeftPadded[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        error: Expression = ...,
        from_: JLeftPadded[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class ExceptionType(Py, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    exception_group: bool
    expression: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        exception_group: bool = ...,
        expression: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class ExpressionStatement(Py, Expression, Statement):
    id: UUID
    expression: Expression

    def replace(
        self,
        *,
        id: UUID = ...,
        expression: Expression = ...,
    ) -> Self: ...

class ExpressionTypeTree(Py, Expression, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    reference: J

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        reference: J = ...,
    ) -> Self: ...

class FormattedString(Py, Expression, TypedTree):
    class Value:
        id: UUID
        prefix: Space
        markers: Markers
        expression: JRightPadded[Expression]
        debug: Optional[Space]
        conversion: Conversion
        format: Optional[Expression]

        def replace(
            self,
            *,
            id: UUID = ...,
            prefix: Space = ...,
            markers: Markers = ...,
            expression: JRightPadded[Expression] = ...,
            debug: Optional[Space] = ...,
            conversion: Conversion = ...,
            format: Optional[Expression] = ...,
        ) -> Self: ...

    id: UUID
    prefix: Space
    markers: Markers
    delimiter: str
    parts: List[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        delimiter: str = ...,
        parts: List[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class KeyValue(Py, Expression, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    key: JRightPadded[Expression]
    value: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        key: JRightPadded[Expression] = ...,
        value: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class LiteralType(Py, Expression, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    literal: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        literal: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class MatchCase(Py, Expression):
    class Pattern:
        id: UUID
        prefix: Space
        markers: Markers
        pattern_kind: Kind
        children: JContainer[Expression]
        type: Optional[Any]

        def replace(
            self,
            *,
            id: UUID = ...,
            prefix: Space = ...,
            markers: Markers = ...,
            pattern_kind: Kind = ...,
            children: JContainer[Expression] = ...,
            type: Optional[Any] = ...,
        ) -> Self: ...

    id: UUID
    prefix: Space
    markers: Markers
    pattern: Pattern
    guard: Optional[JLeftPadded[Expression]]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        pattern: Pattern = ...,
        guard: Optional[JLeftPadded[Expression]] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class MultiImport(Py, Statement):
    id: UUID
    prefix: Space
    markers: Markers
    from_: Optional[JRightPadded[NameTree]]
    parenthesized: bool
    names: JContainer[Statement]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        from_: Optional[JRightPadded[NameTree]] = ...,
        parenthesized: bool = ...,
        names: JContainer[Statement] = ...,
    ) -> Self: ...

class NamedArgument(Py, Expression):
    id: UUID
    prefix: Space
    markers: Markers
    name: J
    value: JLeftPadded[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        name: J = ...,
        value: JLeftPadded[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class Pass(Py, Statement):
    id: UUID
    prefix: Space
    markers: Markers

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
    ) -> Self: ...

class Slice(Py, Expression, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    start: Optional[JRightPadded[Expression]]
    stop: Optional[JRightPadded[Expression]]
    step: Optional[JRightPadded[Expression]]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        start: Optional[JRightPadded[Expression]] = ...,
        stop: Optional[JRightPadded[Expression]] = ...,
        step: Optional[JRightPadded[Expression]] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class SpecialParameter(Py, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    parameter_kind: Kind
    type_hint: Optional[TypeHint]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        parameter_kind: Kind = ...,
        type_hint: Optional[TypeHint] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class Star(Py, Expression, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    star_kind: Kind
    expression: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        star_kind: Kind = ...,
        expression: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class StatementExpression(Py, Expression, Statement):
    id: UUID
    statement: Statement

    def replace(
        self,
        *,
        id: UUID = ...,
        statement: Statement = ...,
    ) -> Self: ...

class TrailingElseWrapper(Py, Statement):
    id: UUID
    prefix: Space
    markers: Markers
    statement: JRightPadded[Statement]
    else_block: J

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        statement: JRightPadded[Statement] = ...,
        else_block: J = ...,
    ) -> Self: ...

class TypeAlias(Py, Statement, TypedTree):
    id: UUID
    prefix: Space
    markers: Markers
    name: J
    type_parameters: Optional[JContainer[J]]
    value: JLeftPadded[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        name: J = ...,
        type_parameters: Optional[JContainer[J]] = ...,
        value: JLeftPadded[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class TypeHint(Py, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    type_tree: Expression
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        type_tree: Expression = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class TypeHintedExpression(Py, Expression):
    id: UUID
    prefix: Space
    markers: Markers
    expression: Expression
    type_hint: TypeHint

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        expression: Expression = ...,
        type_hint: TypeHint = ...,
    ) -> Self: ...

class UnionType(Py, Expression, TypeTree):
    id: UUID
    prefix: Space
    markers: Markers
    types: List[JRightPadded[Expression]]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        types: List[JRightPadded[Expression]] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...

class VariableScope(Py, Statement):
    id: UUID
    prefix: Space
    markers: Markers
    scope_kind: Kind
    names: List[JRightPadded[J]]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        scope_kind: Kind = ...,
        names: List[JRightPadded[J]] = ...,
    ) -> Self: ...

class YieldFrom(Py, Expression):
    id: UUID
    prefix: Space
    markers: Markers
    expression: JLeftPadded[Expression]
    type: Optional[Any]

    def replace(
        self,
        *,
        id: UUID = ...,
        prefix: Space = ...,
        markers: Markers = ...,
        expression: JLeftPadded[Expression] = ...,
        type: Optional[Any] = ...,
    ) -> Self: ...



# Copyright 2025 the original author or authors.
# <p>
# Licensed under the Moderne Source Available License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# <p>
# https://docs.moderne.io/licensing/moderne-source-available-license
# <p>
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Entry point discovery for recipe packages."""

from __future__ import annotations

import inspect
from importlib.metadata import entry_points
from typing import List, Tuple, Type, TYPE_CHECKING

from rewrite.category import CategoryDescriptor
from rewrite.decorators import get_recipe_category
from rewrite.marketplace import RecipeMarketplace
from rewrite.recipe import Recipe

if TYPE_CHECKING:
    pass


def discover_recipes() -> RecipeMarketplace:
    """
    Discover all recipes from installed packages via entry points.

    Looks for packages that declare entry points under
    [project.entry-points."openrewrite.recipes"] and calls their
    activate function to install recipes into the marketplace.

    Returns:
        A RecipeMarketplace containing all discovered recipes.

    Example pyproject.toml entry point:
        [project.entry-points."openrewrite.recipes"]
        mypackage = "mypackage:activate"

    The activate function signature should be:
        def activate(marketplace: RecipeMarketplace) -> None:
            marketplace.install(MyRecipe, [Python, Cleanup])
    """
    marketplace = RecipeMarketplace()

    # Python 3.10+ uses select parameter via SelectableGroups
    eps = entry_points(group="openrewrite.recipes")

    for ep in eps:
        try:
            module = ep.load()
            if hasattr(module, "activate") and callable(module.activate):
                module.activate(marketplace)
        except Exception:
            # Log or handle the error - for now, skip failed activations
            pass

    return marketplace


def discover_decorated_recipes_in_module(
    module,
) -> List[Tuple[Type[Recipe], List[CategoryDescriptor]]]:
    """
    Find all decorated Recipe subclasses in a module.

    This function inspects a module and finds all classes that:
    1. Are subclasses of Recipe
    2. Have been decorated with @recipe

    Args:
        module: A Python module to inspect

    Returns:
        A list of (recipe_class, category_path) tuples for each
        decorated Recipe found in the module.
    """
    results: List[Tuple[Type[Recipe], List[CategoryDescriptor]]] = []

    for _name, obj in inspect.getmembers(module, inspect.isclass):
        # Check if it's a Recipe subclass (but not Recipe itself)
        if issubclass(obj, Recipe) and obj is not Recipe:
            # Check if it has the @recipe decorator
            category = get_recipe_category(obj)
            if category is not None:
                results.append((obj, category))

    return results



import ast
import sys
import token
from argparse import ArgumentError
from functools import lru_cache
from io import BytesIO
from pathlib import Path
from tokenize import tokenize, TokenInfo
from typing import Optional, TypeVar, cast, Callable, List, Tuple, Dict, Type, Sequence, Union, Iterator

from more_itertools import peekable

from rewrite import random_id, Markers, list_map_last
from rewrite.java import Space, JRightPadded, JContainer, JLeftPadded, JavaType, J, Statement, Semicolon, TrailingComma, \
    NameTree, OmitParentheses, Expression, TypeTree, TypedTree, Comment
from rewrite.java import tree as j
from . import tree as py
from .markers import KeywordArguments, KeywordOnlyArguments, Quoted
from .support_types import PyComment
from .type_mapping import PythonTypeMapping

T = TypeVar('T')
J2 = TypeVar('J2', bound=J)


class ParserVisitor(ast.NodeVisitor):
    _source: str
    _cursor: int
    _parentheses_stack: List[Tuple[Callable[[T, Space], T], int, int, ast.AST, Space]]

    @property
    def _source_after_cursor(self) -> str:
        return self._slow_source_after_cursor(self._source, self._cursor)

    @staticmethod
    @lru_cache
    def _slow_source_after_cursor(source: str, cursor: int) -> str:
        return source[cursor:]

    def __init__(self, source: str):
        super().__init__()
        self._source = source
        self._cursor = 0
        self._parentheses_stack = []
        self._type_mapping = PythonTypeMapping(source)

    def generic_visit(self, node):
        return super().generic_visit(node)

    def visit_arguments(self, node, with_close_paren: bool = True) -> List[JRightPadded[j.VariableDeclarations]]:
        first_with_default = len(node.posonlyargs) + len(node.args) - len(
            node.defaults) if node.defaults else sys.maxsize
        if not node.posonlyargs and not node.args and not node.vararg and not node.kwarg and not node.kwonlyargs:
            return [
                JRightPadded(
                    j.Empty(random_id(), self.__source_before(')') if with_close_paren else Space.EMPTY, Markers.EMPTY),
                    Space.EMPTY, Markers.EMPTY)
            ]

        mapped = []
        if node.posonlyargs:
            mapped += [self.__pad_list_element(
                self.map_arg(a, node.defaults[i - first_with_default] if i >= first_with_default else None)) for
                i, a in enumerate(node.posonlyargs)]
            mapped.append(self.__pad_list_element(
                j.VariableDeclarations(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    [],
                    [],
                    None,
                    None,
                    [],
                    [self.__pad_right(
                        j.VariableDeclarations.NamedVariable(
                            random_id(),
                            Space.EMPTY,
                            Markers.EMPTY,
                            cast(j.Identifier, self.__convert_name('/', None)),
                            [],
                            None,
                            None
                        ),
                        Space.EMPTY
                    )]
                ),
                last=not node.args and not node.vararg and not node.kwarg and not node.kwonlyargs,
                end_delim=',' if node.args or node.vararg or node.kwarg or node.kwonlyargs else ')' if with_close_paren else None
            ))

        mapped += [self.__pad_list_element(
            self.map_arg(a, node.defaults[i - first_with_default] if i >= first_with_default else None),
            i == len(node.args) + len(
                node.posonlyargs) - 1 and not node.vararg and not node.kwarg and not node.kwonlyargs,
            end_delim=',' if node.vararg or node.kwarg or node.kwonlyargs else ')' if with_close_paren else None) for
            (i, a) in [(i + len(node.posonlyargs), a) for i, a in enumerate(node.args)]]
        if node.vararg:
            mapped.append(self.__pad_list_element(
                self.map_arg(node.vararg, None, vararg=True),
                not node.kwarg and not node.kwonlyargs,
                end_delim=')' if with_close_paren else None
            ))
        if node.kwonlyargs:
            if not node.vararg:
                empty_name = j.VariableDeclarations.NamedVariable(random_id(), Space.EMPTY, Markers.EMPTY,
                                                                  cast(j.Identifier, self.__convert_name('', None)), [],
                                                                  None, None)
                kwonly_prefix = self.__source_before('*')
                mapped.append(
                    JRightPadded(
                        j.VariableDeclarations(
                            random_id(),
                            kwonly_prefix,
                            Markers(random_id(), [KeywordOnlyArguments(random_id())]),
                            [], [], None, None, [],
                            [self.__pad_right(empty_name, self.__source_before(','))],
                            None
                        ),
                        Space.EMPTY,
                        Markers.EMPTY
                    ))

            for i, kwonlyarg in enumerate(node.kwonlyargs):
                mapped.append(self.__pad_list_element(
                    self.map_arg(kwonlyarg, node.kw_defaults[i], kwarg=False),
                    not node.kwarg and i == len(node.kwonlyargs) - 1,
                    end_delim=')' if with_close_paren else None
                ))

        if node.kwarg:
            mapped.append(self.__pad_list_element(
                self.map_arg(node.kwarg, None, kwarg=True),
                True,
                end_delim=')' if with_close_paren else None
            ))
        return mapped

    def map_arg(self, node, default=None, vararg=False, kwarg=False):
        prefix = self.__source_before('**') if kwarg else self.__whitespace()
        vararg_prefix = self.__source_before('*') if vararg else None
        name = self.__convert_name(node.arg, self._type_mapping.type(node))
        after_name = self.__source_before(':') if node.annotation else Space.EMPTY
        type_expression = self.__convert_type(node.annotation) if node.annotation else None
        initializer = self.__pad_left(self.__source_before('='), self.__convert(default)) if default else None

        return j.VariableDeclarations(
            random_id(),
            prefix,
            Markers(random_id(), [KeywordArguments(random_id())]) if kwarg else Markers.EMPTY,
            [],
            [],
            type_expression,
            vararg_prefix if vararg else None,
            [],
            [self.__pad_right(j.VariableDeclarations.NamedVariable(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                cast(j.Identifier, name),
                [],
                initializer,
                self._type_mapping.type(node)
            ), after_name)],
        )

    def visit_Assert(self, node):
        return j.Assert(
            random_id(),
            self.__source_before('assert'),
            Markers.EMPTY,
            self.__convert(node.test),
            self.__pad_left(self.__source_before(','), self.__convert(node.msg)) if node.msg else None,
        )

    def visit_Assign(self, node):
        if len(node.targets) == 1:
            return j.Assignment(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__convert(node.targets[0]),
                self.__pad_left(self.__source_before('='), self.__convert(node.value)),
                self._type_mapping.type(node.value)
            )
        else:
            return py.ChainedAssignment(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                [self.__pad_list_element(
                    self.__convert(t),
                    i == len(node.targets) - 1,
                    delim='=',
                    end_delim='=',
                    pad_last=True) for i, t in enumerate(node.targets)],
                self.__convert(node.value),
                self._type_mapping.type(node.value)
            )

    def visit_AugAssign(self, node):
        return j.AssignmentOperation(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            self.__convert(node.target),
            self._map_assignment_operator(node.op),
            self.__convert(node.value),
            self._type_mapping.type(node)
        )

    def visit_Await(self, node):
        return py.Await(
            random_id(),
            self.__source_before('await'),
            Markers.EMPTY,
            self.__convert(node.value),
            self._type_mapping.type(node)
        )

    def visit_Interactive(self, node):
        raise NotImplementedError("Implement visit_Interactive!")

    def visit_AsyncFunctionDef(self, node):
        return self.visit_FunctionDef(node)

    def visit_ClassDef(self, node):
        prefix = self.__whitespace()
        decorators = [self.__map_decorator(d) for d in node.decorator_list]
        kind_prefix = self.__source_before('class')
        name = self.__convert_name(node.name)
        save_cursor = self._cursor
        interfaces_prefix = self.__whitespace()
        if (node.bases or node.keywords) and self.__cursor_at('('):
            all = node.bases + node.keywords
            self.__skip('(')
            interfaces = JContainer(
                interfaces_prefix,
                [
                    self.__pad_list_element(self.__convert_type(n), i == len(all) - 1, end_delim=')') for i, n in
                    enumerate(all)],
                Markers.EMPTY
            )
        elif self.__cursor_at('('):
            self.__skip('(')
            interfaces = JContainer(
                interfaces_prefix,
                [self.__pad_right(j.Empty(random_id(), self.__source_before(')'), Markers.EMPTY), Space.EMPTY)],
                Markers.EMPTY
            )
        else:
            interfaces = None
            self._cursor = save_cursor
        return j.ClassDeclaration(
            random_id(),
            prefix,
            Markers.EMPTY,
            decorators,
            [],  # TODO modifiers
            j.ClassDeclaration.Kind(
                random_id(),
                kind_prefix,
                Markers.EMPTY,
                [],
                j.ClassDeclaration.Kind.Type.Class
            ),
            name,
            None,
            None,
            None,  # no `extends`, all in `implements`
            interfaces,
            None,
            self.__convert_block(node.body),
            self._type_mapping.type(node)
        )

    def visit_Delete(self, node):
        return py.Del(
            random_id(),
            self.__source_before('del'),
            Markers.EMPTY,
            [self.__pad_list_element(self.__convert(e), last=i == len(node.targets) - 1) for i, e in
             enumerate(node.targets)]
        )

    def visit_AnnAssign(self, node):
        prefix = self.__whitespace()

        if node.value:
            return j.Assignment(
                random_id(),
                prefix,
                Markers.EMPTY,
                py.TypeHintedExpression(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    self.__convert(node.target),
                    py.TypeHint(
                        random_id(),
                        self.__source_before(':'),
                        Markers.EMPTY,
                        self.__convert_type(node.annotation),
                        self._type_mapping.type(node.annotation)
                    ),
                    self._type_mapping.type(node)
                ),
                self.__pad_left(
                    self.__source_before('='),
                    self.__convert(node.value)
                ) if node.value else None,
                self._type_mapping.type(node)
            )
        elif not node.value:
            return py.ExpressionStatement(
                random_id(),
                py.TypeHintedExpression(
                    random_id(),
                    prefix,
                    Markers.EMPTY,
                    self.__convert(node.target),
                    py.TypeHint(
                        random_id(),
                        self.__source_before(':'),
                        Markers.EMPTY,
                        self.__convert_type(node.annotation),
                        self._type_mapping.type(node.annotation)
                    ),
                    self._type_mapping.type(node)
                )
            )
        else:
            name = cast(j.Identifier, self.__convert(node.target))
            if node.annotation:
                after = self.__source_before(':')
                type = self.__convert_type(node.annotation)
            else:
                after = Space.EMPTY
                type = None

            initializer = self.__pad_left(
                self.__source_before('='),
                self.__convert(node.value)
            ) if node.value else None

            return j.VariableDeclarations(
                random_id(),
                prefix,
                Markers.EMPTY,
                [],
                [],
                type,
                None,
                [],
                [self.__pad_right(
                    j.VariableDeclarations.NamedVariable(
                        random_id(),
                        Space.EMPTY,
                        Markers.EMPTY,
                        name,
                        [],
                        initializer,
                        self._type_mapping.type(node.target)),
                    after
                )]
            )

    def visit_For(self, node):
        prefix = self.__source_before('for')
        target = self.__convert(node.target)
        # Wrap target in ExpressionStatement so it can be used as a Statement
        wrapped_target = py.ExpressionStatement(random_id(), target)
        in_prefix = self.__source_before('in')
        iterable = self.__convert(node.iter)
        body = self.__convert_block(node.body)

        control = j.ForEachLoop.Control(
            random_id(),
            Space.EMPTY,  # No parentheses in Python, so no prefix space for control
            Markers.EMPTY,
            self.__pad_right(wrapped_target, in_prefix),  # Right padding has space before 'in'
            self.__pad_right(iterable, Space.EMPTY)  # ':' comes from body's Block prefix
        )

        loop = j.ForEachLoop(
            random_id(),
            prefix,
            Markers.EMPTY,
            control,
            self.__pad_right(body, Space.EMPTY)
        )

        return loop if not node.orelse else py.TrailingElseWrapper(
            random_id(),
            loop.prefix,
            Markers.EMPTY,
            loop.replace(prefix=Space.EMPTY),
            self.__pad_left(
                self.__source_before('else'),
                self.__convert_block(node.orelse)
            )
        )

    def visit_AsyncFor(self, node):
        return py.Async(
            random_id(),
            self.__source_before('async'),
            Markers.EMPTY,
            self.visit_For(node)
        )

    def visit_While(self, node):
        while_ = j.WhileLoop(
            random_id(),
            self.__source_before('while'),
            Markers.EMPTY,
            j.ControlParentheses(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__pad_right(self.__convert(node.test), Space.EMPTY)
            ),
            self.__pad_right(self.__convert_block(node.body), Space.EMPTY)
        )

        return while_ if not node.orelse else py.TrailingElseWrapper(
            random_id(),
            while_.prefix,
            Markers.EMPTY,
            while_.replace(prefix=Space.EMPTY),
            self.__pad_left(
                self.__source_before('else'),
                self.__convert_block(node.orelse)
            )
        )

    def visit_If(self, node):
        prefix = self.__source_before('if')
        condition = j.ControlParentheses(random_id(), self.__whitespace(), Markers.EMPTY,
                                         self.__pad_right(self.__convert(node.test), Space.EMPTY))
        then = self.__pad_right(self.__convert_block(node.body), Space.EMPTY)
        elze = None
        if len(node.orelse) > 0:
            else_prefix = self.__whitespace()
            if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.If) and self._source.startswith('elif',
                                                                                                        self._cursor):
                is_elif = True
                self._cursor += 2
            else:
                is_elif = False
                self._cursor += 4

            elze = j.If.Else(
                random_id(),
                else_prefix,
                Markers.EMPTY,
                self.__pad_statement(node.orelse[0]) if is_elif else self.__pad_right(
                    self.__convert_block(node.orelse), Space.EMPTY
                )
            )
        return j.If(
            random_id(),
            prefix,
            Markers.EMPTY,
            condition,
            then,
            elze
        )

    def visit_With(self, node):
        prefix = self.__source_before('with')
        items_prefix = self.__whitespace()

        parenthesized = self.__cursor_at('(')
        parens_handler = self.__push_parentheses(node, items_prefix, self._cursor) if parenthesized else None

        resources = [self.__pad_list_element(self.__convert(r), i == len(node.items) - 1) for i, r in
                     enumerate(node.items)]

        if parenthesized and self._parentheses_stack and self._parentheses_stack[-1] is parens_handler:
            self._cursor += 1
            resources_container = self._parentheses_stack.pop()[0](
                JContainer(items_prefix, resources, Markers.EMPTY),
                Space.EMPTY
            )
        else:
            resources_container = JContainer(
                items_prefix if not parenthesized else Space.EMPTY,
                resources,
                Markers.build(random_id(), [OmitParentheses(random_id())])
            )

        return j.Try(
            random_id(),
            prefix,
            Markers.EMPTY,
            resources_container,
            self.__convert_block(node.body),
            [],
            None
        )

    def visit_withitem(self, node):
        prefix = self.__whitespace()
        expr = self.__convert(node.context_expr)
        if node.optional_vars:
            value = self.__pad_left(self.__source_before('as'), expr)
            name = self.__convert(node.optional_vars)
            var = j.Assignment(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                name,
                value,
                self._type_mapping.type(node.context_expr)
            )
        else:
            var = expr

        if not isinstance(var, TypedTree):
            var = py.ExpressionTypeTree(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                var
            )

        return j.Try.Resource(
            random_id(),
            prefix,
            Markers.EMPTY,
            var,
            False
        )

    def visit_AsyncWith(self, node):
        return py.Async(
            random_id(),
            self.__source_before('async'),
            Markers.EMPTY,
            self.visit_With(node)
        )

    def visit_Raise(self, node):
        prefix = self.__source_before('raise')
        if node.cause:
            exc = py.ErrorFrom(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__convert(node.exc),
                self.__pad_left(self.__source_before('from'), self.__convert(node.cause)),
                self._type_mapping.type(node)
            )
        elif node.exc:
            exc = self.__convert(node.exc)
        else:
            exc = j.Empty(random_id(), Space.EMPTY, Markers.EMPTY)

        return j.Throw(
            random_id(),
            prefix,
            Markers.EMPTY,
            exc,
        )

    def visit_Try(self, node):
        prefix = self.__source_before('try')
        body = self.__convert_block(node.body)
        handlers = [cast(j.Try.Catch, self.__convert(handler)) for handler in node.handlers]
        if node.orelse:
            else_block = self.__pad_left(
                self.__source_before('else'),
                self.__convert_block(node.orelse)
            )

        finally_ = self.__pad_left(self.__source_before('finally'),
                                   self.__convert_block(node.finalbody)) if node.finalbody else None
        try_ = j.Try(random_id(), prefix, Markers.EMPTY, JContainer.empty(), body, handlers, finally_)

        return try_ if not node.orelse else py.TrailingElseWrapper(
            random_id(),
            try_.prefix,
            Markers.EMPTY,
            try_.replace(prefix=Space.EMPTY),
            else_block
        )

    def visit_Import(self, node):
        # TODO only use `MultiImport` when necessary (requires corresponding changes to printer)
        return py.MultiImport(
            random_id(),
            self.__source_before('import'),
            Markers.EMPTY,
            None,
            False,
            JContainer(
                Space.EMPTY,
                [self.__pad_list_element(self.__convert(n), i == len(node.names) - 1, pad_last=False) for i, n in
                 enumerate(node.names)],
                Markers.EMPTY
            )
        )

    def visit_ImportFrom(self, node):
        prefix = self.__source_before('from')
        from_ = self.__pad_right(self.__convert_name(('.' * node.level) + (node.module if node.module else '')),
                                 self.__source_before('import'))
        names_prefix = self.__whitespace()
        if parenthesized := self._source[self._cursor] == '(':
            self.__skip('(')
        multi_import = py.MultiImport(
            random_id(),
            prefix,
            Markers.EMPTY,
            from_,
            parenthesized,
            JContainer(
                names_prefix,
                [self.__pad_list_element(self.__convert(n), i == len(node.names) - 1) for i, n in
                 enumerate(node.names)],
                Markers.EMPTY
            )
        )
        if parenthesized:
            self.__skip(')')
        return multi_import

    def visit_alias(self, node):
        return j.Import(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            self.__pad_left(Space.EMPTY, False),
            self.__convert_qualified_name(node.name),
            None if not node.asname else
            self.__pad_left(self.__source_before('as'), self.__convert_name(node.asname))
        )

    def visit_keyword(self, node):
        if node.arg:
            return py.NamedArgument(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__convert_name(node.arg),
                self.__pad_left(self.__source_before('='), self.__convert(node.value)),
                self._type_mapping.type(node)
            )
        prefix = self.__whitespace()
        if self._source.startswith('**', self._cursor):
            self.__skip('**')
            return py.Star(
                random_id(),
                prefix,
                Markers.EMPTY,
                py.Star.Kind.DICT,
                self.__convert(node.value),
                self._type_mapping.type(node.value),
            )

    def __convert_qualified_name(self, name: str) -> j.FieldAccess:
        if '.' not in name:
            return j.FieldAccess(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                j.Empty(random_id(), Space.EMPTY, Markers.EMPTY),
                self.__pad_left(
                    Space.EMPTY,
                    self.__convert_name(name)
                ),
                None
            )
        return cast(j.FieldAccess, self.__convert_name(name))

    def visit_Global(self, node):
        return py.VariableScope(
            random_id(),
            self.__source_before('global'),
            Markers.EMPTY,
            py.VariableScope.Kind.GLOBAL,
            [self.__pad_list_element(
                cast(j.Identifier, self.__convert_name(n)),
                i == len(node.names) - 1,
                pad_last=False) for i, n in enumerate(node.names)
            ]
        )

    def visit_Nonlocal(self, node):
        return py.VariableScope(
            random_id(),
            self.__source_before('nonlocal'),
            Markers.EMPTY,
            py.VariableScope.Kind.NONLOCAL,
            [self.__pad_list_element(
                cast(j.Identifier, self.__convert_name(n)),
                i == len(node.names) - 1,
                pad_last=False) for i, n in enumerate(node.names)
            ]
        )

    def visit_Pass(self, node):
        return py.Pass(
            random_id(),
            self.__source_before('pass'),
            Markers.EMPTY,
        )

    def visit_Break(self, node):
        return j.Break(random_id(), self.__source_before('break'), Markers.EMPTY, None)

    def visit_Continue(self, node):
        return j.Continue(random_id(), self.__source_before('continue'), Markers.EMPTY, None)

    def visit_GeneratorExp(self, node):
        # this weird logic is here to deal with the case of generator expressions appearing as the argument to a call
        prefix = self.__whitespace()
        if self._source[self._cursor] == '(':
            save_cursor = self._cursor
            self._cursor += 1
            try:
                result = self.__convert(node.elt)
                save_cursor_2 = self._cursor
                self.__whitespace()
                assert self._source[self._cursor] != ')'
                self._cursor = save_cursor_2
                parenthesized = True
            except:
                self._cursor = save_cursor
                result = self.__convert(node.elt)
                parenthesized = False
        else:
            result = self.__convert(node.elt)
            parenthesized = False

        return py.ComprehensionExpression(
            random_id(),
            prefix,
            Markers.EMPTY if parenthesized else Markers.EMPTY.replace(markers=[OmitParentheses(random_id())]),
            py.ComprehensionExpression.Kind.GENERATOR,
            result,
            cast(List[py.ComprehensionExpression.Clause], [self.__convert(g) for g in node.generators]),
            self.__source_before(')') if parenthesized else Space.EMPTY,
            self._type_mapping.type(node)
        )

    def visit_Expr(self, node):
        return self.__convert(node.value)

    def visit_Yield(self, node):
        return py.StatementExpression(
            random_id(),
            j.Yield(
                random_id(),
                self.__source_before('yield'),
                Markers.EMPTY,
                False,
                self.__convert(node.value) if node.value else j.Empty(random_id(), Space.EMPTY, Markers.EMPTY),
            )
        )

    def visit_YieldFrom(self, node):
        return py.StatementExpression(
            random_id(),
            j.Yield(
                random_id(),
                self.__source_before('yield'),
                Markers.EMPTY,
                False,
                py.YieldFrom(
                    random_id(),
                    self.__source_before('from'),
                    Markers.EMPTY,
                    self.__convert(node.value),
                    self._type_mapping.type(node)
                )
            )
        )

    def visit_TypeIgnore(self, node):
        raise NotImplementedError("Implement visit_TypeIgnore!")

    def visit_Attribute(self, node):
        return j.FieldAccess(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            self.__convert(node.value),
            self.__pad_left(self.__source_before('.'), self.__convert_name(node.attr)),
            self._type_mapping.type(node),
        )

    def visit_Del(self, node):
        raise NotImplementedError("Implement visit_Del!")

    def visit_Load(self, node):
        raise NotImplementedError("Implement visit_Load!")

    def visit_Store(self, node):
        raise NotImplementedError("Implement visit_Store!")

    def visit_ExceptHandler(self, node):
        prefix = self.__source_before('except')
        type_prefix = self.__whitespace()
        except_type = self.__convert_type(node.type) if node.type else j.Empty(random_id(), Space.EMPTY,
                                                                               Markers.EMPTY)
        if node.name:
            before_as = self.__source_before('as')
            except_type_name = self.__convert_name(node.name)
        else:
            before_as = Space.EMPTY
            except_type_name = j.Identifier(random_id(), Space.EMPTY, Markers.EMPTY, [], '', None, None)
        except_type_name = self.__pad_right(
            j.VariableDeclarations.NamedVariable(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                except_type_name,
                [], None, None
            ),
            before_as
        )

        return j.Try.Catch(
            random_id(),
            prefix,
            Markers.EMPTY,
            j.ControlParentheses(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                self.__pad_right(j.VariableDeclarations(
                    random_id(),
                    type_prefix,
                    Markers.EMPTY,
                    [], [],
                    except_type,
                    None, [],
                    [except_type_name]
                ), Space.EMPTY)
            ),
            self.__convert_block(node.body)
        )

    def visit_Match(self, node):
        return j.Switch(
            random_id(),
            self.__source_before('match'),
            Markers.EMPTY,
            j.ControlParentheses(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__pad_right(self.__convert(node.subject), Space.EMPTY)
            ),
            self.__convert_block(node.cases)
        )

    def visit_match_case(self, node):
        prefix = self.__source_before('case')
        pattern_prefix = self.__whitespace()

        pattern = self.__convert(node.pattern)
        if isinstance(pattern, py.MatchCase) and node.guard:
            guard = self.__pad_left(self.__source_before('if'), self.__convert(node.guard))
            pattern = pattern.padding.replace(guard=guard)

        return j.Case(
            random_id(),
            prefix,
            Markers.EMPTY,
            j.Case.Type.Rule,
            JContainer(
                pattern_prefix,
                [self.__pad_right(pattern, Space.EMPTY)],
                Markers.EMPTY
            ),
            JContainer.empty(),
            self.__pad_right(self.__convert_block(node.body), Space.EMPTY),
            None
        )

    def visit_MatchValue(self, node):
        return self.__convert(node.value)

    def visit_MatchSequence(self, node):
        prefix = self.__whitespace()
        end_delim = None
        if self._source[self._cursor] == '[':
            kind = py.MatchCase.Pattern.Kind.SEQUENCE_LIST
            self._cursor += 1
            end_delim = ']'
        elif self._source[self._cursor] == '(':
            kind = py.MatchCase.Pattern.Kind.SEQUENCE_TUPLE
            self._cursor += 1
            end_delim = ')'
        else:
            kind = py.MatchCase.Pattern.Kind.SEQUENCE
        return py.MatchCase(
            random_id(),
            Space.EMPTY,
            Markers.EMPTY,
            py.MatchCase.Pattern(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                kind,
                JContainer(
                    prefix,
                    [self.__pad_list_element(self.__convert(e), last=i == len(node.patterns) - 1,
                                             end_delim=end_delim) for i, e in
                     enumerate(node.patterns)] if node.patterns else [],
                    Markers.EMPTY
                ),
                None
            ),
            None,
            None
        )

    def visit_MatchSingleton(self, node):
        raise NotImplementedError("Implement visit_MatchSingleton!")

    def visit_MatchStar(self, node):
        return py.Star(
            random_id(),
            self.__source_before('*'),
            Markers.EMPTY,
            py.Star.Kind.LIST,
            self.__convert_name(node.name),
            None
        )

    def visit_MatchMapping(self, node):
        return py.MatchCase(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            py.MatchCase.Pattern(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                py.MatchCase.Pattern.Kind.MAPPING,
                JContainer(
                    self.__source_before('{'),
                    [self.__pad_list_element(py.MatchCase.Pattern(
                        random_id(),
                        Space.EMPTY,
                        Markers.EMPTY,
                        py.MatchCase.Pattern.Kind.KEY_VALUE,
                        JContainer(
                            Space.EMPTY,
                            [
                                self.__pad_right(self.__convert(node.keys[i]), self.__source_before(':')),
                                self.__pad_right(self.__convert(node.patterns[i]), Space.EMPTY),
                            ],
                            Markers.EMPTY
                        ),
                        None
                    ), last=i == len(node.patterns) - 1, end_delim='}') for i, e in
                        enumerate(node.patterns)] if node.patterns else [],
                    Markers.EMPTY
                ),
                None
            ),
            None,
            None
        )

    def visit_MatchClass(self, node):
        prefix = self.__whitespace()
        children = [self.__pad_right(self.__convert(node.cls), self.__source_before('('))]
        if len(node.patterns) > 0:
            for i, arg in enumerate(node.patterns):
                arg_name = j.VariableDeclarations(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    [], [], None, None, [],
                    [
                        self.__pad_right(j.VariableDeclarations.NamedVariable(
                            random_id(),
                            Space.EMPTY,
                            Markers.EMPTY,
                            cast(j.Identifier, self.__convert(arg)),
                            [],
                            None,
                            None
                        ), Space.EMPTY)
                    ]
                )
                converted = self.__pad_list_element(arg_name, last=i == len(node.patterns) - 1,
                                                    end_delim=')' if len(node.kwd_attrs) == 0 else ',')
                children.append(converted)
            for i, kwd in enumerate(node.kwd_attrs):
                kwd_var = j.VariableDeclarations(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    [], [], None, None, [],
                    [
                        self.__pad_right(j.VariableDeclarations.NamedVariable(
                            random_id(),
                            Space.EMPTY,
                            Markers.EMPTY,
                            cast(j.Identifier, self.__convert_name(kwd)),
                            [],
                            self.__pad_left(self.__source_before('='), self.__convert(node.kwd_patterns[i])),
                            None
                        ), Space.EMPTY)
                    ]
                )
                converted = self.__pad_list_element(kwd_var, last=i == len(node.kwd_attrs) - 1,
                                                    end_delim=')')
                children.append(converted)
        else:
            children.append(
                self.__pad_right(j.Empty(random_id(), self.__source_before(')'), Markers.EMPTY), Space.EMPTY))
        return py.MatchCase(
            random_id(),
            prefix,
            Markers.EMPTY,
            py.MatchCase.Pattern(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                py.MatchCase.Pattern.Kind.CLASS,
                JContainer(Space.EMPTY, children, Markers.EMPTY),
                None
            ),
            None,
            None
        )

    def visit_MatchAs(self, node):
        if node.name is None and node.pattern is None:
            return py.MatchCase(
                random_id(),
                self.__source_before('_'),
                Markers.EMPTY,
                py.MatchCase.Pattern(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    py.MatchCase.Pattern.Kind.WILDCARD,
                    JContainer.empty(),
                    None
                ),
                None,
                None
            )
        elif node.pattern:
            return py.MatchCase(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                py.MatchCase.Pattern(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    py.MatchCase.Pattern.Kind.AS,
                    JContainer(
                        Space.EMPTY,
                        [
                            self.__pad_right(self.__convert(node.pattern), self.__source_before('as')),
                            self.__pad_right(self.__convert_name(node.name), Space.EMPTY),
                        ],
                        Markers.EMPTY
                    ),
                    None
                ),
                None,
                None
            )
        return self.__convert_name(node.name)

    def visit_MatchOr(self, node):
        return py.MatchCase(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            py.MatchCase.Pattern(
                random_id(),
                Space.EMPTY,
                Markers.EMPTY,
                py.MatchCase.Pattern.Kind.OR,
                JContainer(
                    Space.EMPTY,
                    [self.__pad_list_element(self.__convert(e), last=i == len(node.patterns) - 1) for i, e in
                     enumerate(node.patterns)] if node.patterns else [],
                    Markers.EMPTY
                ),
                None
            ),
            None,
            None
        )

    def visit_TryStar(self, node):
        raise NotImplementedError("Implement visit_TryStar!")

    def visit_TypeVar(self, node):
        raise NotImplementedError("Implement visit_TypeVar!")

    def visit_ParamSpec(self, node):
        raise NotImplementedError("Implement visit_ParamSpec!")

    def visit_TypeVarTuple(self, node):
        raise NotImplementedError("Implement visit_TypeVarTuple!")

    def visit_TypeAlias(self, node):
        return py.TypeAlias(
            random_id(),
            self.__source_before("type"),
            Markers.EMPTY,
            self.__convert(node.name),
            self.__pad_left(self.__source_before('='), self.__convert(node.value)),
            self._type_mapping.type(node)
        )

    def visit_ExtSlice(self, node):
        raise NotImplementedError("Implement visit_ExtSlice!")

    def visit_Index(self, node):
        raise NotImplementedError("Implement visit_Index!")

    def visit_Suite(self, node):
        raise NotImplementedError("Implement visit_Suite!")

    def visit_AugLoad(self, node):
        raise NotImplementedError("Implement visit_AugLoad!")

    def visit_AugStore(self, node):
        raise NotImplementedError("Implement visit_AugStore!")

    def visit_Param(self, node):
        raise NotImplementedError("Implement visit_Param!")

    def visit_Num(self, node):
        raise NotImplementedError("Implement visit_Num!")

    def visit_Str(self, node):
        raise NotImplementedError("Implement visit_Str!")

    def visit_Bytes(self, node):
        raise NotImplementedError("Implement visit_Bytes!")

    def visit_NameConstant(self, node):
        raise NotImplementedError("Implement visit_NameConstant!")

    def visit_Ellipsis(self, node):
        raise NotImplementedError("Implement visit_Ellipsis!")

    def visit_BinOp(self, node):
        prefix = self.__whitespace()
        left = self.__convert(node.left)
        op = self.__convert_binary_operator(node.op)

        if isinstance(op.element, py.Binary.Type):
            return py.Binary(
                random_id(),
                prefix,
                Markers.EMPTY,
                left,
                op,
                None,
                self.__convert(node.right),
                self._type_mapping.type(node)
            )
        else:
            return j.Binary(
                random_id(),
                prefix,
                Markers.EMPTY,
                left,
                op,
                self.__convert(node.right),
                self._type_mapping.type(node)
            )

    def visit_BoolOp(self, node):
        binaries = []
        prefix = Space.EMPTY
        left = self.__convert(node.values[0])
        for right_expr in node.values[1:]:
            left = j.Binary(
                random_id(),
                prefix,
                Markers.EMPTY,
                left,
                self.__convert_binary_operator(node.op),
                self.__convert(right_expr),
                self._type_mapping.type(node)
            )
            binaries.append(left)
            prefix = Space.EMPTY

        return binaries[-1]

    def visit_Call(self, node):
        prefix = self.__whitespace()
        if isinstance(node.func, ast.Name):
            select = None
            name = cast(j.Identifier, self.__convert(node.func))
        elif isinstance(node.func, ast.Attribute):
            select = self.__pad_right(self.__convert(node.func.value), self.__source_before('.'))
            name = j.Identifier(
                random_id(),
                self.__source_before(node.func.attr),
                Markers.EMPTY,
                [],
                node.func.attr,
                self._type_mapping.type(node),
                None
            )
            save_cursor = self._cursor
            parens_right_padding = self.__whitespace()
            if self.__cursor_at(')'):
                name = j.FieldAccess(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    select.element,
                    JLeftPadded(select.after, name, Markers.EMPTY),
                    None
                )
                while len(self._parentheses_stack) > 0 and self.__cursor_at(')'):
                    self._cursor += 1
                    name = self._parentheses_stack.pop()[0](name, parens_right_padding)
                    save_cursor = self._cursor
                    parens_right_padding = self.__whitespace()
            self._cursor = save_cursor
        else:
            select = self.__pad_right(cast(Expression, self.__convert(node.func)), self.__whitespace())
            # printer handles empty name by not printing `.` before it
            name = self.__convert_name('')

        all_args = self.__sort_call_arguments(node)
        args = JContainer(
            self.__source_before('('),
            [self.__pad_list_element(self.__convert(a), last=i == len(all_args) - 1, end_delim=')') for i, a in
             enumerate(all_args)] if all_args else [
                self.__pad_right(j.Empty(random_id(), self.__source_before(')'), Markers.EMPTY),
                                 Space.EMPTY)],
            Markers.EMPTY
        )

        return j.MethodInvocation(
            random_id(),
            prefix,
            Markers.EMPTY,
            select if isinstance(name, j.Identifier) else self.__pad_right(name, Space.EMPTY),
            None,
            name if isinstance(name, j.Identifier) else j.Identifier(random_id(), Space.EMPTY, Markers.EMPTY, [], "",
                                                                     None, None),
            args,
            name.type if isinstance(name.type, JavaType.Method) else None,
        )

    def __sort_call_arguments(self, call: ast.Call) -> List[Union[ast.expr, ast.keyword]]:
        all_args = []

        for arg in call.args:
            all_args.append((arg.lineno, arg.col_offset, arg))

        for kw in call.keywords:
            all_args.append((kw.value.lineno, kw.value.col_offset, kw))

        all_args.sort(key=lambda x: (x[0], x[1]))
        return [arg[2] for arg in all_args]

    def visit_Compare(self, node):
        prefix = self.__whitespace()
        left = self.__convert(node.left)

        for i in range(len(node.ops)):
            op = self.__convert_binary_operator(node.ops[i])

            if isinstance(op.element, j.Binary.Type):
                left = j.Binary(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    left,
                    op,
                    self.__convert(node.comparators[i]),
                    self._type_mapping.type(node)
                )
            else:
                if op.element == py.Binary.Type.IsNot:
                    negation = self.__source_before('not')
                elif op.element == py.Binary.Type.NotIn:
                    negation = self.__source_before('in')
                else:
                    negation = None

                left = py.Binary(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    left,
                    op,
                    negation,
                    self.__convert(node.comparators[i]),
                    self._type_mapping.type(node)
                )

        return left.replace(prefix=prefix)

    def __convert_binary_operator(self, op) -> Union[JLeftPadded[j.Binary.Type], JLeftPadded[py.Binary.Type]]:
        operation_map: Dict[Type[ast], Tuple[j.Binary.Type, str]] = {
            ast.Add: (j.Binary.Type.Addition, '+'),
            ast.And: (j.Binary.Type.And, 'and'),
            ast.BitAnd: (j.Binary.Type.BitAnd, '&'),
            ast.BitOr: (j.Binary.Type.BitOr, '|'),
            ast.BitXor: (j.Binary.Type.BitXor, '^'),
            ast.Div: (j.Binary.Type.Division, '/'),
            ast.Eq: (j.Binary.Type.Equal, '=='),
            ast.FloorDiv: (py.Binary.Type.FloorDivision, '//'),
            ast.Gt: (j.Binary.Type.GreaterThan, '>'),
            ast.GtE: (j.Binary.Type.GreaterThanOrEqual, '>='),
            ast.In: (py.Binary.Type.In, 'in'),
            ast.Is: (py.Binary.Type.Is, 'is'),
            ast.IsNot: (py.Binary.Type.IsNot, 'is'),
            ast.LShift: (j.Binary.Type.LeftShift, '<<'),
            ast.Lt: (j.Binary.Type.LessThan, '<'),
            ast.LtE: (j.Binary.Type.LessThanOrEqual, '<='),
            ast.MatMult: (py.Binary.Type.MatrixMultiplication, '@'),
            ast.Mod: (j.Binary.Type.Modulo, '%'),
            ast.Mult: (j.Binary.Type.Multiplication, '*'),
            ast.NotEq: (j.Binary.Type.NotEqual, '!='),
            ast.NotIn: (py.Binary.Type.NotIn, 'not'),
            ast.Or: (j.Binary.Type.Or, 'or'),
            ast.Pow: (py.Binary.Type.Power, '**'),
            ast.RShift: (j.Binary.Type.RightShift, '>>'),
            ast.Sub: (j.Binary.Type.Subtraction, '-'),
        }
        try:
            op, op_str = operation_map[type(op)]
        except KeyError:
            raise ValueError(f"Unsupported operator: {op}")
        return self.__pad_left(self.__source_before(op_str), op)

    def visit_Constant(self, node):
        tokens = peekable(tokenize(BytesIO(self._source[self._cursor:].encode('utf-8')).readline))
        tok = next(tokens)  # skip ENCODING token
        tok = next(tokens)  # skip ENCODING token
        while tok.type in (token.ENCODING, token.NL, token.NEWLINE, token.INDENT, token.DEDENT, token.COMMENT):
            tok = next(tokens)

        if not isinstance(node.value, (str, bytes)):
            return self.__map_literal(node, tok, tokens)[0]

        is_byte_string = tok.string.startswith(('b', "B"))
        res = None
        end_seen = False

        while tok.type in (token.STRING, token.FSTRING_START) and is_byte_string == tok.string.startswith(('b', "B")):
            if tok.type == token.FSTRING_START:
                prefix = self.__whitespace()
                current, tok, _ = self.__map_fstring(node, prefix, tok, tokens)
            else:
                current, tok = self.__map_literal(node, tok, tokens)
                end_seen = current.value.endswith(node.value)

            if res is None:
                res = current
            else:
                res = py.Binary(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    res,
                    self.__pad_left(Space.EMPTY, py.Binary.Type.StringConcatenation),
                    None,
                    current,
                    self._type_mapping.type(node)
                )

            if end_seen:
                break

            while tok.type in (token.NL, token.NEWLINE, token.INDENT, token.DEDENT, token.COMMENT):
                try:
                    tok = next(tokens)
                except IndentationError:
                    break

        return res

    def __map_literal(self, node, tok, tokens):
        prefix = self.__whitespace()
        start = self._cursor
        self._cursor += len(tok.string)

        return (j.Literal(
            random_id(),
            prefix,
            Markers.EMPTY,
            self.__map_literal_value(node, tok),
            self._source[start:self._cursor],
            None,
            self._type_mapping.type(node),
        ), next(tokens))

    def __map_literal_value(self, node, tok):
        if node.value is Ellipsis:
            return None
        elif isinstance(node.value, (str, bytes)):
            return ast.literal_eval(ast.parse(tok.string, mode='eval').body)
        return node.value

    def visit_Dict(self, node):
        return py.DictLiteral(
            random_id(),
            self.__source_before('{'),
            Markers.EMPTY,
            JContainer(
                Space.EMPTY,
                [self.__pad_right(j.Empty(random_id(), self.__source_before('}'), Markers.EMPTY),
                                  Space.EMPTY)] if not node.keys else
                [self.__map_dict_entry(k, v, i == len(node.keys) - 1) for i, (k, v) in
                 enumerate(zip(node.keys, node.values))],
                Markers.EMPTY
            ),
            self._type_mapping.type(node)
        )

    def visit_DictComp(self, node):
        return py.ComprehensionExpression(
            random_id(),
            self.__source_before('{'),
            Markers.EMPTY,
            py.ComprehensionExpression.Kind.DICT,
            py.KeyValue(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__pad_right(self.__convert(node.key), self.__source_before(':')),
                self.__convert(node.value),
                self._type_mapping.type(node.value)
            ),
            cast(List[py.ComprehensionExpression.Clause], [self.__convert(g) for g in node.generators]),
            self.__source_before('}'),
            self._type_mapping.type(node)
        )

    def __map_dict_entry(self, key: Optional[ast.expr], value: ast.expr, last: bool) -> JRightPadded[J]:
        if key is None:
            element = py.Star(
                random_id(),
                self.__source_before('**'),
                Markers.EMPTY,
                py.Star.Kind.DICT,
                self.__convert(value),
                self._type_mapping.type(value),
            )
        else:
            element = py.KeyValue(random_id(), self.__whitespace(), Markers.EMPTY,
                                  self.__pad_right(self.__convert(key), self.__source_before(':')),
                                  self.__convert(value),
                                  self._type_mapping.type(value))
        return self.__pad_list_element(element, last, end_delim='}')

    def visit_FunctionDef(self, node: ast.FunctionDef) -> j.MethodDeclaration:
        prefix = self.__whitespace()
        decorators = [self.__map_decorator(d) for d in node.decorator_list]

        modifiers = []
        if isinstance(node, ast.AsyncFunctionDef):
            modifiers.append(j.Modifier(
                random_id(),
                self.__source_before('async'),
                Markers.EMPTY,
                'async',
                j.Modifier.Type.Async,
                []
            ))

        def_prefix = self.__source_before('def')
        modifiers.append(j.Modifier(
            random_id(),
            def_prefix,
            Markers.EMPTY,
            'def',
            j.Modifier.Type.Default,
            []
        ))
        name = j.MethodDeclaration.IdentifierWithAnnotations(j.Identifier(
            random_id(),
            self.__source_before(node.name),
            Markers.EMPTY,
            [],
            node.name,
            None,
            None
        ), [])

        params = JContainer(self.__source_before('('), self.visit_arguments(node.args), Markers.EMPTY)
        if node.returns is None:
            return_type = None
        else:
            return_type = py.TypeHint(
                random_id(),
                self.__source_before('->'),
                Markers.EMPTY,
                self.__convert(node.returns),
                self._type_mapping.type(node.returns)
            )
        body = self.__convert_block(node.body)

        return j.MethodDeclaration(
            random_id(),
            prefix,
            Markers.EMPTY,
            decorators,
            modifiers,
            None,
            return_type,
            name,
            params,
            None,
            body,
            None,
            self._type_mapping.type(node),
        )

    def __map_decorator(self, decorator) -> j.Annotation:
        prefix = self.__source_before('@')
        if isinstance(decorator, (ast.Attribute, ast.Name)):
            name = self.__convert(decorator)
            args = None
        elif isinstance(decorator, ast.Call):
            name = self.__convert(decorator.func)
            all_args = decorator.args + decorator.keywords
            args = JContainer(
                self.__source_before('('),
                [self.__pad_right(j.Empty(random_id(), self.__source_before(')'), Markers.EMPTY),
                                  Space.EMPTY)] if not all_args else
                [self.__pad_list_element(self.__convert(a), i == len(all_args) - 1, end_delim=')') for i, a in
                 enumerate(all_args)],
                Markers.EMPTY
            )
        else:
            raise NotImplementedError("Unsupported decorator type: " + str(type(decorator)))

        return j.Annotation(
            random_id(),
            prefix,
            Markers.EMPTY,
            name,
            args
        )

    def visit_IfExp(self, node):
        # TODO check if we actually want to use `J.Ternary` as it requires "reversing" some of the padding
        prefix = self.__whitespace()
        true_expr = self.__convert(node.body)
        true_part = self.__pad_left(self.__source_before('if'), true_expr)
        condition = self.__convert(node.test)
        false_part = self.__pad_left(self.__source_before('else'), self.__convert(node.orelse))
        return j.Ternary(
            random_id(),
            prefix,
            Markers.EMPTY,
            condition,
            true_part,
            false_part,
            self._type_mapping.type(node)
        )

    def visit_JoinedStr(self, node):
        tokens = peekable(tokenize(BytesIO(self._source[self._cursor:].encode('utf-8')).readline))
        next(tokens)  # skip ENCODING token
        while (tok := next(tokens)).type not in (token.FSTRING_START, token.STRING):
            pass

        value_idx = 0
        res = None
        while tok.type in (token.FSTRING_START, token.STRING):
            if tok.type == token.STRING:
                current, tok = self.__map_literal(node.values[value_idx], tok, tokens)
                if current.value.endswith(node.values[value_idx].value):
                    value_idx += 1
            else:
                prefix = self.__whitespace()
                current, tok, value_idx = self.__map_fstring(node, prefix, tok, tokens, value_idx)

            if res is None:
                res = current
            else:
                res = py.Binary(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    res,
                    self.__pad_left(Space.EMPTY, py.Binary.Type.StringConcatenation),
                    None,
                    current,
                    self._type_mapping.type(node)
                )

            if value_idx >= len(node.values):
                break

            while tok.type in (token.NL, token.NEWLINE, token.INDENT, token.DEDENT, token.COMMENT):
                try:
                    tok = next(tokens)
                except IndentationError:
                    break

        return res

    def visit_FormattedValue(self, node):
        raise ValueError("This method should not be called directly")

    def visit_Lambda(self, node):
        return j.Lambda(
            random_id(),
            self.__source_before('lambda'),
            Markers.EMPTY,
            j.Lambda.Parameters(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                False,
                self.visit_arguments(node.args, with_close_paren=False)
            ),
            self.__source_before(':'),
            self.__convert(node.body),
            self._type_mapping.type(node)
        )

    def visit_List(self, node):
        prefix = self.__source_before('[')
        elements = JContainer(
            Space.EMPTY,
            [self.__pad_list_element(self.__convert(e), last=i == len(node.elts) - 1, end_delim=']') for i, e in
             enumerate(node.elts)] if node.elts else
            [self.__pad_right(j.Empty(random_id(), self.__source_before(']'), Markers.EMPTY), Space.EMPTY)],
            Markers.EMPTY
        )
        return py.CollectionLiteral(
            random_id(),
            prefix,
            Markers.EMPTY,
            py.CollectionLiteral.Kind.LIST,
            elements,
            self._type_mapping.type(node)
        )

    def visit_ListComp(self, node):
        return py.ComprehensionExpression(
            random_id(),
            self.__source_before('['),
            Markers.EMPTY,
            py.ComprehensionExpression.Kind.LIST,
            self.__convert(node.elt),
            cast(List[py.ComprehensionExpression.Clause], [self.__convert(g) for g in node.generators]),
            self.__source_before(']'),
            self._type_mapping.type(node)
        )

    def visit_comprehension(self, node):
        if node.is_async:
            prefix = self.__source_before('async')
            async_ = JRightPadded(True, self.__source_before('for'), Markers.EMPTY)
        else:
            prefix = self.__source_before('for')
            async_ = None

        return py.ComprehensionExpression.Clause(
            random_id(),
            prefix,
            Markers.EMPTY,
            async_,
            self.__convert(node.target),
            self.__pad_left(self.__source_before('in'), self.__convert(node.iter)),
            [self._map_comprehension_condition(i) for i in node.ifs] if node.ifs else []
        )

    def _map_comprehension_condition(self, i):
        return py.ComprehensionExpression.Condition(
            random_id(),
            self.__source_before('if'),
            Markers.EMPTY,
            self.__convert(i)
        )

    def visit_Module(self, node: ast.Module) -> py.CompilationUnit:
        self._type_mapping.resolve_types(node)

        cu = py.CompilationUnit(
            random_id(),
            Space.EMPTY,
            Markers.EMPTY,
            Path("TODO"),
            None,
            None,
            False,
            None,
            [],
            [self.__pad_statement(stmt) for stmt in node.body] if node.body else [
                self.__pad_right(j.Empty(random_id(), Space.EMPTY, Markers.EMPTY), Space.EMPTY)],
            self.__whitespace()
        )
        # assert self._cursor == len(self._source)
        return cu

    def visit_Name(self, node):
        return j.Identifier(
            random_id(),
            self.__source_before(node.id),
            Markers.EMPTY,
            [],
            node.id,
            self._type_mapping.type(node),
            None
        )

    def visit_NamedExpr(self, node):
        return j.Assignment(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            self.__convert(node.target),
            self.__pad_left(self.__source_before(':='), self.__convert(node.value)),
            self._type_mapping.type(node.value)
        )

    def visit_Return(self, node):
        return j.Return(
            random_id(),
            self.__source_before('return'),
            Markers.EMPTY,
            self.__convert(node.value) if node.value else None
        )

    def visit_Set(self, node):
        prefix = self.__source_before('{')
        elements = JContainer(
            Space.EMPTY,
            [self.__pad_list_element(self.__convert(e), last=i == len(node.elts) - 1, end_delim='}') for i, e in
             enumerate(node.elts)] if node.elts else
            [self.__pad_right(j.Empty(random_id(), self.__source_before('}'), Markers.EMPTY), Space.EMPTY)],
            Markers.EMPTY
        )
        return py.CollectionLiteral(
            random_id(),
            prefix,
            Markers.EMPTY,
            py.CollectionLiteral.Kind.SET,
            elements,
            self._type_mapping.type(node)
        )

    def visit_SetComp(self, node):
        return py.ComprehensionExpression(
            random_id(),
            self.__source_before('{'),
            Markers.EMPTY,
            py.ComprehensionExpression.Kind.SET,
            self.__convert(node.elt),
            cast(List[py.ComprehensionExpression.Clause], [self.__convert(g) for g in node.generators]),
            self.__source_before('}'),
            self._type_mapping.type(node)
        )

    def visit_Slice(self, node):
        prefix = self.__whitespace()

        lower_expr = self.__convert(node.lower) if node.lower else None
        right_padding = self.__whitespace(':') if node.lower else Space.EMPTY
        lower = self.__pad_right(lower_expr, right_padding) if lower_expr else None
        self._cursor += 1

        upper_expr = self.__convert(node.upper) if node.upper else None
        right_padding = self.__whitespace()
        upper = self.__pad_right(upper_expr if node.upper else j.Empty(random_id(), Space.EMPTY, Markers.EMPTY), right_padding)
        has_step = self.__cursor_at(':')
        if has_step:
            self._cursor += 1
            step = self.__pad_right(
                self.__convert(node.step) if node.step else j.Empty(random_id(), Space.EMPTY, Markers.EMPTY),
                self.__whitespace(']')) if node.step or has_step else Space.EMPTY
        else:
            step = None

        return py.Slice(
            random_id(),
            prefix,
            Markers.EMPTY,
            lower,
            upper,
            step
        )

    def visit_Starred(self, node):
        return py.Star(
            random_id(),
            self.__source_before('*'),
            Markers.EMPTY,
            py.Star.Kind.LIST,
            self.__convert(node.value),
            self._type_mapping.type(node),
        )

    def visit_Subscript(self, node):
        return j.ArrayAccess(
            random_id(),
            self.__whitespace(),
            Markers.EMPTY,
            self.__convert(node.value),
            j.ArrayDimension(
                random_id(),
                self.__source_before('['),
                Markers.EMPTY,
                self.__pad_right(self.__convert(node.slice), self.__source_before(']'))
            ),
            self._type_mapping.type(node)
        )

    def visit_Tuple(self, node):
        prefix = self.__whitespace()
        save_cursor = self._cursor
        maybe_parens = self.__cursor_at('(')

        if maybe_parens:
            self._cursor += 1
            omit_parens = False
        else:
            self._cursor = save_cursor
            omit_parens = True

        elements = JContainer(
            Space.EMPTY,
            [self.__pad_list_element(self.__convert(e), last=i == len(node.elts) - 1) for i, e in enumerate(node.elts)],
            Markers.EMPTY
        ) if node.elts else JContainer(
            Space.EMPTY,
            [self.__pad_right(j.Empty(random_id(), self.__whitespace(), Markers.EMPTY), Space.EMPTY)],
            Markers.EMPTY
        )

        if len(self._parentheses_stack) > 0 and self.__cursor_at(')'):
            self._cursor += 1
            elements = self._parentheses_stack.pop()[0](elements, Space.EMPTY)
            omit_parens = False
        elif maybe_parens and len(self._parentheses_stack) > 0 and self._parentheses_stack[-1][3] == node:
            elements = self._parentheses_stack.pop()
        else:
            if elements.padding.elements:
                # remove the last element's right padding so that if becomes the prefix of the next LST element
                padding = elements.padding.elements[-1].after
                padding_len = len(padding.whitespace) if padding.whitespace else 0
                for c in padding.comments:
                    padding_len += len(c.text) + 1 + len(c.suffix)
                unpadded_last = elements.padding.elements[-1].replace(after=Space.EMPTY)
                elements = elements.padding.replace(elements=list_map_last(lambda last: unpadded_last, elements.padding.elements))
                self._cursor -= padding_len
            omit_parens = True

        return py.CollectionLiteral(
            random_id(),
            prefix,
            Markers.EMPTY,
            py.CollectionLiteral.Kind.TUPLE,
            elements.replace(markers=
                Markers.build(random_id(), [OmitParentheses(random_id())])) if omit_parens else elements,
            self._type_mapping.type(node)
        )

    def visit_UnaryOp(self, node):
        mapped = self._map_unary_operator(node.op)
        return j.Unary(
            random_id(),
            self.__source_before(mapped[1]),
            Markers.EMPTY,
            self.__pad_left(Space.EMPTY, mapped[0]),
            self.__convert(node.operand),
            self._type_mapping.type(node)
        )

    def __convert_type(self, node) -> Optional[TypeTree]:
        prefix = self.__whitespace()
        converted_type = self.__convert_internal(node, self.__convert_type, self.__convert_type_mapper)
        if isinstance(converted_type, TypeTree):
            return converted_type.replace(prefix=prefix)
        else:
            return py.ExpressionTypeTree(
                random_id(),
                prefix,
                Markers.EMPTY,
                converted_type
            )

    def __convert_type_mapper(self, node) -> Optional[TypeTree]:
        if isinstance(node, ast.Constant):
            if node.value is None or node.value is Ellipsis:
                return py.LiteralType(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    self.__convert(node),
                    self._type_mapping.type(node)
                )
            else:
                literal = cast(j.Literal, self.__convert(node))
                if literal.value_source.startswith("'''"):
                    quote_style = Quoted.Style.TRIPLE_SINGLE
                elif literal.value_source[0] == "'":
                    quote_style = Quoted.Style.SINGLE
                elif literal.value_source.startswith('"""'):
                    quote_style = Quoted.Style.TRIPLE_DOUBLE
                elif literal.value_source.startswith('"'):
                    quote_style = Quoted.Style.DOUBLE
                else:
                    quote_style = None

                return j.Identifier(
                    random_id(),
                    literal.prefix,
                    Markers.build(random_id(), [Quoted(random_id(), quote_style)]) if quote_style else Markers.EMPTY,
                    [],
                    str(literal.value),
                    self._type_mapping.type(node),
                    None
                )
        elif isinstance(node, ast.Subscript):
            slices = node.slice.elts if isinstance(node.slice, ast.Tuple) and node.slice.elts else [node.slice]
            return j.ParameterizedType(
                random_id(),
                self.__whitespace(),
                Markers.EMPTY,
                self.__convert(node.value),
                JContainer(
                    self.__source_before('['),
                    [self.__pad_list_element(self.__convert_type(s), last=i == len(slices) - 1, end_delim=']') for
                     i, s in
                     enumerate(slices)],
                    Markers.EMPTY
                ),
                None
            )
        elif isinstance(node, ast.BinOp):
            # NOTE: Type unions using `|` was added in Python 3.10
            prefix = self.__whitespace()
            # FIXME consider flattening nested unions
            left = self.__pad_right(self.__convert_internal(node.left, self.__convert_type),
                                    self.__source_before('|'))
            right = self.__pad_right(self.__convert_internal(node.right, self.__convert_type), Space.EMPTY)
            return py.UnionType(
                random_id(),
                prefix,
                Markers.EMPTY,
                [left, right],
                self._type_mapping.type(node)
            )

        return self.__convert_internal(node, self.__convert_type)

    def __convert(self, node) -> Optional[J]:
        return self.__convert_internal(node, self.__convert)

    def __convert_statement(self, node) -> Optional[J]:
        converted = self.__convert_internal(node, self.__convert_statement)
        if isinstance(converted, Statement):
            return converted
        return py.ExpressionStatement(
            random_id(),
            converted
        )

    def __convert_internal(self, node, recursion, mapping = None) -> Optional[J]:
        if not node or not isinstance(node, ast.expr) or isinstance(node, ast.GeneratorExp):
            return self.visit(cast(ast.AST, node)) if node else None

        save_cursor = self._cursor
        prefix = self.__whitespace()

        # Handle normal expression or parenthesized expression
        result = self.__parse_expr(node, mapping or self.visit, recursion, save_cursor, prefix)

        save_cursor_2 = self._cursor
        suffix = self.__whitespace()

        # Process closing parentheses if any
        while self._is_closing_paren(save_cursor):
            self._cursor += 1
            transformer, save_cursor, _, _, _ = self._parentheses_stack.pop()
            result = transformer(result, suffix)
            save_cursor_2 = self._cursor
            suffix = self.__whitespace()

        # Clean up unmatched parentheses for this node
        while len(self._parentheses_stack) > 1 and self._parentheses_stack[-1][3] is node and \
                self._parentheses_stack[-2][3] is not node:
            self._parentheses_stack.pop()

        self._cursor = save_cursor_2
        return result

    def __parse_expr(self, node, mapping, recursion, save_cursor: int, prefix: Space) -> J:
        """Parse either a normal expression or a parenthesized expression."""
        if not (self._cursor < len(self._source) and self._source[self._cursor] == '('):
            self._cursor = save_cursor
            return mapping(cast(ast.AST, node))

        self.__push_parentheses(node, prefix, save_cursor)

        return recursion(node)

    def __push_parentheses(self, node, prefix: Space, save_cursor):
        self._cursor += 1
        expr_prefix = self.__whitespace()
        handler = (
            lambda e, r: (
                (lambda c: c.padding.replace(elements=[
                    padded.replace(element=padded.element.replace(prefix=expr_prefix)) if i == 0 else padded
                    for i, padded in enumerate(c.padding.elements)
                ]))(cast(JContainer, e).replace(before=prefix))
                if isinstance(e, JContainer)
                else j.Parentheses(
                    random_id(),
                    prefix,
                    Markers.EMPTY,
                    self.__pad_right(e.replace(prefix=expr_prefix), r)
                )
            ),
            save_cursor,
            self._cursor,
            node,
            prefix
        )
        self._parentheses_stack.append(handler)
        return handler

    def _is_closing_paren(self, save_cursor: int) -> bool:
        """Check if current position has a valid closing parenthesis."""
        if (not self._parentheses_stack or
                self._cursor >= len(self._source) or
                self._source[self._cursor] != ')'):
            return False

        stack_cursor = self._parentheses_stack[-1][2]
        slice_content = self._source[stack_cursor:save_cursor]
        return stack_cursor == save_cursor or slice_content.isspace() or slice_content == '('

    def __convert_name(self, name: str, name_type: Optional[JavaType] = None) -> NameTree:
        def ident_or_field(parts: List[str]) -> NameTree:
            if len(parts) == 1:
                return j.Identifier(random_id(), self.__source_before(parts[-1]), Markers.EMPTY, [], parts[-1],
                                    name_type, None)
            else:
                return j.FieldAccess(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    ident_or_field(parts[:-1]),
                    self.__pad_left(
                        self.__source_before('.'),
                        j.Identifier(random_id(), self.__source_before(parts[-1]), Markers.EMPTY, [], parts[-1],
                                     name_type,
                                     None),
                    ),
                    name_type
                )

        return ident_or_field(name.split('.'))

    def __next_lexer_token(self, tokens: Iterator[TokenInfo]) -> TokenInfo:
        tok = next(tokens)
        value_source = tok.string
        self._cursor += len(value_source)
        return tok

    def __convert_all(self, trees: Sequence) -> List[J2]:
        return [self.__convert(tree) for tree in trees]

    def __convert_block(self, statements: Sequence[Statement], prefix: str = ':') -> j.Block:
        prefix = self.__source_before(prefix)
        if statements:
            statements = [self.__pad_statement(cast(ast.stmt, s)) for s in statements]
        else:
            statements = [self.__pad_right(j.Empty(random_id(), Space.EMPTY, Markers.EMPTY), Space.EMPTY)]
        return j.Block(
            random_id(),
            prefix,
            Markers.EMPTY,
            JRightPadded(False, Space.EMPTY, Markers.EMPTY),
            statements,
            Space.EMPTY
        )

    def __pad_statement(self, stmt: ast.stmt) -> JRightPadded[Statement]:
        statement = self.__convert_statement(stmt)
        # use whitespace until end of line as padding; what follows will be the prefix of next element
        # This includes trailing comments on the same line
        padding = self.__whitespace('\n')
        if self._cursor < len(self._source) and self._source[self._cursor] == ';':
            self._cursor += 1
            markers = Markers.build(random_id(), [Semicolon(random_id())])
        else:
            markers = Markers.EMPTY
        return JRightPadded(statement, padding, markers)

    def __pad_list_element(self, element: J2, last: bool = False, pad_last: bool = True, delim: str = ',',
                           end_delim: Optional[str] = None) -> JRightPadded[J2]:
        save_cursor = self._cursor
        padding = self.__whitespace() if pad_last or not last else Space.EMPTY
        markers = Markers.EMPTY
        if last and self._cursor < len(self._source):
            if self._source[self._cursor] == delim and end_delim != delim:
                self._cursor += len(delim)
                markers = markers.replace(markers=[TrailingComma(random_id(), self.__whitespace())])
            elif self._source[self._cursor] != end_delim:
                if not pad_last:
                    self._cursor = save_cursor
            if end_delim and self._source[self._cursor] == end_delim:
                self._cursor += len(end_delim)
        elif last:
            padding = Space.EMPTY
            self._cursor = save_cursor
        elif not last:
            self._cursor += len(delim)
            markers = Markers.EMPTY
        return JRightPadded(element, padding, markers)

    def __pad_right(self, tree, space: Space) -> JRightPadded[J2]:
        return JRightPadded(tree, space, Markers.EMPTY)

    def __pad_left(self, space: Space, tree) -> JLeftPadded[J2]:
        if isinstance(tree, ast.AST):
            raise ArgumentError(tree, "must be a Tree but is a {}".format(type(tree)))
        return JLeftPadded(space, tree, Markers.EMPTY)

    def __source_before(self, until_delim: Optional[str], stop: Optional[str] = None) -> Space:
        if not until_delim:
            return Space.EMPTY

        if self._source.startswith(until_delim, self._cursor):
            self._cursor += len(until_delim)
            return Space.EMPTY

        save_cursor = self._cursor
        space = self.__whitespace()
        if self._source.startswith(until_delim, self._cursor):
            self._cursor += len(until_delim)
            return space
        else:
            self._cursor = save_cursor

        delim_index = self.__position_of_next(until_delim, stop)
        if delim_index == -1:
            return Space.EMPTY

        if delim_index == self._cursor:
            self._cursor += len(until_delim)
            return Space.EMPTY

        space = self.__whitespace()
        self._cursor = delim_index + len(until_delim)
        return space

    def __skip(self, tok: Optional[str]) -> Optional[str]:
        if tok is None:
            return None
        if self._source.startswith(tok, self._cursor):
            self._cursor += len(tok)
        return tok

    def __whitespace(self, stop: Optional[str] = None) -> Space:
        space, self._cursor = self.__format(self._source, self._cursor, stop)
        return space

    def __format(self, source: str, offset: int, stop: Optional[str] = None) -> Tuple[Space, int]:
        prefix = None
        whitespace = []
        comments: List[Comment] = []
        source_len = len(source)
        while offset < source_len:
            char = source[offset]
            if stop is not None and char == stop:
                break
            if char.isspace() or char == '\\':
                whitespace.append(char)
            elif char == '#':
                if comments:
                    comments[-1] = comments[-1].replace(suffix=''.join(whitespace))
                else:
                    prefix = ''.join(whitespace)
                whitespace = []
                comment = []
                offset += 1
                while offset < source_len and source[offset] not in ['\r', '\n']:
                    comment.append(source[offset])
                    offset += 1
                comments.append(PyComment(''.join(comment), '',
                                          False, Markers.EMPTY))
                continue
            else:
                break
            if offset < source_len:
                offset += 1

        if not whitespace and not comments:
            return Space.EMPTY, offset

        if not comments:
            prefix = ''.join(whitespace)
        elif whitespace:
            comments[-1] = comments[-1].replace(suffix=''.join(whitespace))
        return Space(comments, prefix), offset

    def __position_of_next(self, until_delim: str, stop: Optional[str] = None) -> int:
        in_single_line_comment = False

        delim_index = self._cursor
        end_index = len(self._source) - len(until_delim) + 1
        while delim_index < end_index:
            if in_single_line_comment:
                if self._source[delim_index] == '\n':
                    in_single_line_comment = False
            else:
                if self._source[delim_index] == '#':
                    in_single_line_comment = True

            if not in_single_line_comment:
                if stop is not None and self._source.startswith(stop, delim_index):
                    return -1  # reached stop word before finding the delimiter

                if self._source.startswith(until_delim, delim_index):
                    break  # found it!

            delim_index += 1

        return -1 if delim_index > len(self._source) - len(until_delim) else delim_index

    def _map_unary_operator(self, op) -> Tuple[j.Unary.Type, str]:
        operation_map: Dict[Type[ast], Tuple[j.Unary.Type, str]] = {
            ast.Invert: (j.Unary.Type.Complement, '~'),
            ast.Not: (j.Unary.Type.Not, 'not'),
            ast.UAdd: (j.Unary.Type.Positive, '+'),
            ast.USub: (j.Unary.Type.Negative, '-'),
        }
        return operation_map[type(op)]

    def _map_assignment_operator(self, op):
        operation_map: Dict[Type[ast], Tuple[j.AssignmentOperation.Type, str]] = {
            ast.Add: (j.AssignmentOperation.Type.Addition, '+='),
            ast.BitAnd: (j.AssignmentOperation.Type.BitAnd, '&='),
            ast.BitOr: (j.AssignmentOperation.Type.BitOr, '|='),
            ast.BitXor: (j.AssignmentOperation.Type.BitXor, '^='),
            ast.Div: (j.AssignmentOperation.Type.Division, '/='),
            ast.Pow: (j.AssignmentOperation.Type.Exponentiation, '**='),
            ast.FloorDiv: (j.AssignmentOperation.Type.FloorDivision, '//='),
            ast.LShift: (j.AssignmentOperation.Type.LeftShift, '<<='),
            ast.MatMult: (j.AssignmentOperation.Type.MatrixMultiplication, '@='),
            ast.Mod: (j.AssignmentOperation.Type.Modulo, '%='),
            ast.Mult: (j.AssignmentOperation.Type.Multiplication, '*='),
            ast.RShift: (j.AssignmentOperation.Type.RightShift, '>>='),
            ast.Sub: (j.AssignmentOperation.Type.Subtraction, '-='),
        }
        try:
            op, op_str = operation_map[type(op)]
        except KeyError:
            raise ValueError(f"Unsupported operator: {op}")
        return self.__pad_left(self.__source_before(op_str), op)

    def __map_fstring(self, node: ast.JoinedStr, prefix: Space, tok: TokenInfo, tokens: peekable, value_idx: int = 0) -> \
            Tuple[J, TokenInfo, int]:
        if tok.type != token.FSTRING_START:
            if len(node.values) == 1 and isinstance(node.values[0], ast.Constant):
                # format specifiers are stored as f-strings in the AST; e.g. `f'{1:n}'`
                format = cast(ast.Constant, node.values[0]).value
                self._cursor += len(format)
                return (j.Literal(
                    random_id(),
                    self.__whitespace(),
                    Markers.EMPTY,
                    format,
                    format,
                    None,
                    JavaType.Primitive.String
                ), next(tokens), 0)
            else:
                delimiter = ''
            consume_end_delim = False
        else:
            delimiter = tok.string
            self._cursor += len(delimiter)
            tok = next(tokens)
            consume_end_delim = True

        # tokenizer tokens: FSTRING_START, FSTRING_MIDDLE, OP, ..., OP, FSTRING_MIDDLE, FSTRING_END
        parts = []
        while tok.type != token.FSTRING_END and value_idx < len(node.values):
            value = node.values[value_idx]
            if tok.type == token.FSTRING_MIDDLE:
                save_cursor = self._cursor
                while True:
                    self._cursor += len(tok.string) + (1 if tok.string.endswith('{') or tok.string.endswith('}') else 0)
                    if (tok := next(tokens)).type != token.FSTRING_MIDDLE:
                        break
                s = self._source[save_cursor:self._cursor]
                parts.append(j.Literal(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    s,
                    s,
                    None,
                    JavaType.Primitive.String
                ))
                if cast(ast.Constant, value).value == s:
                    value_idx += 1
            elif tok.type == token.OP and tok.string == '{':
                self._cursor += 1
                tok = next(tokens)
                if not isinstance(value, ast.FormattedValue):
                    # this is the case when using the `=` "debug specifier"
                    value_idx += 1
                    value = node.values[value_idx]

                if isinstance(cast(ast.FormattedValue, value).value, ast.JoinedStr):
                    nested, tok, _ = self.__map_fstring(cast(ast.JoinedStr, cast(ast.FormattedValue, value).value),
                                                        Space.EMPTY, tok, tokens)
                    expr = self.__pad_right(
                        nested,
                        Space.EMPTY
                    )
                else:
                    expr = self.__pad_right(
                        self.__convert(cast(ast.FormattedValue, value).value),
                        self.__whitespace()
                    )
                    try:
                        while (tokens.peek()).type not in (token.FSTRING_END, token.FSTRING_MIDDLE):
                            tok = next(tokens)
                            if tok.type == token.OP and tok.string in ('!'):
                                break
                            la_tok = tokens.peek()
                            if tok.type == token.OP and tok.string == '}' and (
                                    la_tok.type in (token.FSTRING_END, token.FSTRING_MIDDLE) or (
                                    la_tok.type == token.OP and la_tok.string == '{')):
                                break
                            if tok.type == token.OP and tok.string == '=' and la_tok.string in ('!', ':', '}'):
                                break
                            if tok.type == token.OP and tok.string == ':' and la_tok.string in ('{'):
                                break
                    except StopIteration:
                        pass

                # debug specifier
                if tok.type == token.OP and tok.string == '=':
                    self._cursor += len(tok.string)
                    tok = next(tokens)
                    debug = self.__pad_right(True, self.__whitespace('\n'))
                else:
                    debug = None

                # conversion specifier
                if tok.type == token.OP and tok.string == '!':
                    self._cursor += len(tok.string)
                    tok = next(tokens)
                    conv = py.FormattedString.Value.Conversion.ASCII if tok.string == 'a' else py.FormattedString.Value.Conversion.STR if tok.string == 's' else py.FormattedString.Value.Conversion.REPR
                    self._cursor += len(tok.string)
                    tok = next(tokens)
                else:
                    conv = None

                # format specifier
                if tok.type == token.OP and tok.string == ':':
                    self._cursor += len(tok.string)
                    format_spec, tok, _ = self.__map_fstring(
                        cast(ast.JoinedStr, cast(ast.FormattedValue, value).format_spec), Space.EMPTY, next(tokens),
                        tokens)
                else:
                    format_spec = None

                parts.append(py.FormattedString.Value(
                    random_id(),
                    Space.EMPTY,
                    Markers.EMPTY,
                    expr,
                    debug,
                    conv,
                    format_spec
                ))
                value_idx += 1
                self._cursor += len(tok.string)
                tok = next(tokens)
            elif tok.type == token.FSTRING_END:
                raise NotImplementedError("Unsupported: String concatenation with f-strings")

        if consume_end_delim:
            self._cursor += len(tok.string)  # FSTRING_END token
            tok = next(tokens)
        elif tok.type == token.FSTRING_MIDDLE and len(tok.string) == 0:
            tok = next(tokens)

        return (py.FormattedString(
            random_id(),
            prefix,
            Markers.EMPTY,
            delimiter,
            parts
        ), tok, value_idx)

    def __cursor_at(self, s: str):
        return self._cursor < len(self._source) and (
                len(s) == 1 and self._source[self._cursor] == s or self._source.startswith(s, self._cursor))
